# SPDX-License-Identifier: BSD-2-Clause
""" Provides details of the RTEMS specification. """

# Copyright (C) 2021, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
import functools
import hashlib
import logging
import os
from typing import Callable, Iterable

from specitems import (EnabledSet, Item, ItemGetValueContext, ItemMapper,
                       is_enabled, Link)
from specware import (augment_with_test_case_links, augment_with_test_links,
                      get_items_by_type_map, get_item_types_by_prefix,
                      get_items_by_types, get_interface_items,
                      get_interface_and_requirement_items,
                      get_requirement_items, is_validation_by_test,
                      recursive_is_enabled, validate)

from .pkgitems import (BuildItem, BuildItemFactory, build_item_input,
                       GenericPackageComponent, PackageBuildDirector,
                       PackageComponent)


def _add_unique_name(name_to_item: dict[str, Item], item: Item,
                     name: str) -> None:
    if name in name_to_item:
        raise ValueError(
            f"items {item.uid} and {name_to_item[name].uid} "
            f"specify the same interface group: {name.partition('/')[2]}")
    name_to_item[name] = item


def _name_default(name_to_item: dict[str, Item], item: Item) -> None:
    name_to_item[item.ident] = item


def _name_define(name_to_item: dict[str, Item], item: Item) -> None:
    if not item.enabled:
        return
    _name_default(name_to_item, item)
    name = f"define/{item['name']}"
    if name in name_to_item:
        if not name_to_item[name].type.startswith("interface/unspecified"):
            if not item.type.startswith("interface/unspecified"):
                raise ValueError(f"duplicate defines specified by {item.uid} "
                                 f"and {name_to_item[name].uid}")
            return
    name_to_item[name] = item


def _name_kind(kind: str, key: str, name_to_item: dict[str, Item],
               item: Item) -> None:
    if not item.enabled:
        return
    _name_default(name_to_item, item)
    _add_unique_name(name_to_item, item, f"{kind}/{item[key]}")


_name_enum = functools.partial(_name_kind, "enum", "name")
_name_enumerator = functools.partial(_name_kind, "enumerator", "name")
_name_function = functools.partial(_name_kind, "function", "name")
_name_group = functools.partial(_name_kind, "group", "identifier")
_name_object = functools.partial(_name_kind, "object", "name")
_name_type = functools.partial(_name_kind, "type", "name")
_name_perf_runtime_env = functools.partial(_name_kind, "perf-runtime-env",
                                           "name")


def _name_design_group(name_to_item: dict[str, Item], item: Item) -> None:
    if not item.enabled:
        return
    _name_default(name_to_item, item)
    identifier = item["identifier"]
    if identifier is None:
        return
    _add_unique_name(name_to_item, item, f"group/{identifier}")


def _name_ident_group(name_to_item: dict[str, Item], item: Item) -> None:
    if not item.enabled:
        return
    _name_default(name_to_item, item)
    _add_unique_name(name_to_item, item, f"group/{item.ident}")


def _name_header(name_to_item: dict[str, Item], item: Item) -> None:
    if not item.enabled:
        return
    _name_default(name_to_item, item)
    name = item.get("prefix", "")
    if name:
        name = f"file/{name}/{item['path']}"
    else:
        name = f"file/{item['path']}"
    _add_unique_name(name_to_item, item, name)


def _name_appl_config_group(name_to_item: dict[str, Item], item: Item) -> None:
    if not item.enabled:
        return
    _name_default(name_to_item, item)
    _add_unique_name(name_to_item, item,
                     f"group/RTEMSApplConfig{item['name'].replace(' ', '')}")


def _name_test_program(name_to_item: dict[str, Item], item: Item) -> None:
    _name_default(name_to_item, item)
    _add_unique_name(name_to_item, item,
                     f"file/{os.path.basename(item['target'])}")


def _name_script(name_to_item: dict[str, Item], item: Item) -> None:
    if item.get("target", None):
        _name_test_program(name_to_item, item)


_NAME: dict[str, Callable[[dict[str, Item], Item], None]] = {
    "build/script": _name_script,
    "build/test-program": _name_test_program,
    "interface/appl-config-group": _name_appl_config_group,
    "interface/appl-config-option/feature": _name_define,
    "interface/appl-config-option/feature-enable": _name_define,
    "interface/appl-config-option/initializer": _name_define,
    "interface/appl-config-option/integer": _name_define,
    "interface/define": _name_define,
    "interface/enum": _name_enum,
    "interface/enumerator": _name_enumerator,
    "interface/function": _name_function,
    "interface/group": _name_group,
    "interface/header-file": _name_header,
    "interface/macro": _name_default,
    "interface/register-block": _name_group,
    "interface/struct": _name_type,
    "interface/typedef": _name_type,
    "interface/union": _name_type,
    "interface/unspecified-define": _name_define,
    "interface/unspecified-enum": _name_enum,
    "interface/unspecified-enumerator": _name_enumerator,
    "interface/unspecified-function": _name_function,
    "interface/unspecified-header-file": _name_header,
    "interface/unspecified-macro": _name_define,
    "interface/unspecified-object": _name_object,
    "interface/unspecified-struct": _name_type,
    "interface/unspecified-typedef": _name_type,
    "interface/unspecified-union": _name_type,
    "interface/variable": _name_object,
    "memory-benchmark": _name_ident_group,
    "requirement/functional/action": _name_ident_group,
    "requirement/non-functional/design-group": _name_design_group,
    "requirement/non-functional/performance-runtime-environment":
    _name_perf_runtime_env,
    "runtime-measurement-test": _name_ident_group,
    "test-case": _name_ident_group,
    "test-suite": _name_ident_group,
}


def _get_issue(ctx: ItemGetValueContext) -> str:
    database = ctx.item.parent("issue-member")
    mapper = ItemMapper(ctx.item)
    url = mapper.substitute(database["url"])
    identifier = mapper.substitute(database["format-identifier"])
    return f"`{database['name']} {identifier} <{url}>`__"


class RTEMSItemCache(BuildItem):
    """ Augments the items with RTEMS-specific attributes and links. """

    @classmethod
    def prepare_factory(cls, factory: BuildItemFactory,
                        type_name: str) -> None:
        BuildItem.prepare_factory(factory, type_name)
        factory.add_get_value("pkg/issue:/name", _get_issue)

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self.item_cache = self.item.cache
        logging.info("%s: augment with test links", self.uid)
        augment_with_test_links(self.item_cache)
        logging.info("%s: augment with test case links", self.uid)
        augment_with_test_case_links(self.item_cache)
        logging.info("%s: gather name information", self.uid)
        self.name_to_item: dict[str, Item] = {}
        for item_2 in self.item_cache.values():
            _NAME.get(item_2.type, _name_default)(self.name_to_item, item_2)
        self.related_items: set[Item] = set()
        self.related_items_by_type: dict[str, list[Item]] = {}
        self.related_validations_by_test: set[Item] = set()

        # Calculate the overall item cache hash.  Ignore package configuration
        # items and specification type changes.
        state = hashlib.sha512()
        for item_2 in sorted(self.item_cache.values()):
            if not item_2.type.startswith(("pkg", "spec")):
                digest = item_2.get("workspace-digest", None)
                if digest is None:
                    digest = item_2.digest
                state.update(digest.encode("ascii"))
        self._hash = base64.urlsafe_b64encode(state.digest()).decode("ascii")

    def run(self) -> None:
        self.description.add("""Augment the component with an RTEMS-specific
view of the specification items.""")

    def _validate_using_test_results(self, item: Item,
                                     validated: bool) -> bool:
        if is_validation_by_test(item):
            self.related_validations_by_test.add(item)

            # A validation by test is a valid validation, if it has at least
            # one test result and the test did not unexpectedly fail (it passed
            # or is an expected failure) in all test results.
            results = False
            for test_results in item.view.get("test-results", {}).values():
                for data in test_results:
                    if data["status"] == "F":
                        return False
                    results = True
            if not results:
                return False
        elif item.type == "requirement/non-functional/design-target":
            if not item.view.get("no-unexpected-test-failures", False):
                return False
        return validated

    def validate_using_test_results(self) -> None:
        """ Validate the specification using test results. """
        self.related_items = validate(self.item_cache[self["spec-root-uid"]],
                                      self._validate_using_test_results)
        self.related_items_by_type = get_items_by_type_map(self.related_items)

    def has_changed(self, link: Link) -> bool:
        return link["hash"] is None or self._hash != link["hash"]

    def refresh_link(self, link: Link) -> None:
        link["hash"] = self._hash

    def get_related_items_by_type(self,
                                  types: str | Iterable[str]) -> list[Item]:
        """ Get related items by a list of types. """
        return get_items_by_types(self.related_items_by_type, types)

    def get_related_types_by_prefix(
        self,
        prefix: str | tuple[str, ...],
        exclude: tuple[str, ...] = tuple()
    ) -> list[str]:
        """
        Get the types of the related items having one of the type prefixes.
        """
        return get_item_types_by_prefix(self.related_items_by_type, prefix,
                                        exclude)

    def get_related_interfaces(self) -> list[Item]:
        """ Get the related interfaces. """
        return get_interface_items(self.related_items_by_type)

    def get_related_requirements(self) -> list[Item]:
        """ Get the related requirements. """
        return get_requirement_items(self.related_items_by_type)

    def get_related_interfaces_and_requirements(self) -> list[Item]:
        """ Get the related interfaces and requirements. """
        return get_interface_and_requirement_items(self.related_items_by_type)


class RTEMSPackageComponent(GenericPackageComponent):
    """ Provides a package with an RTEMS-specific item selection.  """

    def run(self) -> None:
        self.description.add("""Provide settings for the RTEMS-specific
component.""")

    def is_item_enabled(self, enabled_set: EnabledSet, item: Item) -> bool:
        try:
            component_item = build_item_input(item, "component")
        except KeyError:
            return recursive_is_enabled(enabled_set, item)
        component = self.director[component_item.uid]
        assert isinstance(component, PackageComponent)
        return is_enabled(component.selection.enabled_set, item["enabled-by"])



# SPDX-License-Identifier: BSD-2-Clause
"""
Provides functions for the generation of memory benchmark documentation.
"""

# Copyright (C) 2021, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import re
from typing import Any, NamedTuple

from specitems import (COL_SPAN, Item, ItemCache, ItemMapper, SphinxContent,
                       get_reference)
from specware import run_command


class MembenchVariant(NamedTuple):
    """ Represents a static memory benchmark configuration. """
    name: str
    build_label: str


SectionsByUID = dict[str, dict[str, int]]

_SECTIONS = {".text": 0, ".rodata": 1, ".data": 2, ".bss": 3, ".noinit": 4}

_SECTION_KEYS = tuple(sorted(_SECTIONS.keys(), key=lambda x: _SECTIONS[x]))

_SECTION = re.compile(
    r"^\s*\d+\s+(\S+)\s+([0-9a-fA-F]+)\s+([0-9a-fA-F]+)\s+[0-9a-fA-F]+"
    r"\s+[0-9a-fA-F]+\s+.*$")

_SECTION_MAP = {
    ".ARM.attributes": None,
    ".ARM.exidx": ".rodata",
    ".PPC.EMB.apuinfo": None,
    ".bss": ".bss",
    ".comment": None,
    ".ctors": ".rodata",
    ".data": ".data",
    ".data.rel.ro": ".rodata",
    ".exceptions": ".text",
    ".toc": ".rodata",
    ".debug_abbrev": None,
    ".debug_aranges": None,
    ".debug_frame": None,
    ".debug_gdb_scripts": None,
    ".debug_info": None,
    ".debug_line": None,
    ".debug_line_str": None,
    ".debug_loc": None,
    ".debug_loclists": None,
    ".debug_pubnames": None,
    ".debug_ranges": None,
    ".debug_rnglists": None,
    ".debug_str": None,
    ".dtors": ".rodata",
    ".eh_frame": ".rodata",
    ".eh_frame_hdr": ".rodata",
    ".fast_data": ".data",
    ".fast_text": ".text",
    ".fini": ".text",
    ".fini_array": ".rodata",
    ".gnu.attributes": None,
    ".got": ".rodata",
    ".got2": ".rodata",
    ".htif": ".data",
    ".init": ".text",
    ".init_array": ".rodata",
    ".nocachenoload": None,
    ".noinit": ".noinit",
    ".riscv.attributes": None,
    ".robarrier": None,
    ".rodata": ".rodata",
    ".rtemsroset": ".rodata",
    ".rtemsrwset": ".data",
    ".rtemsstack": ".noinit",
    ".rwbarrier": None,
    ".sbss": ".bss",
    ".sdata": ".data",
    ".stack": None,
    ".start": ".text",
    ".startdata": ".text",
    ".text": ".text",
    ".vector": ".vector",
    ".work": None,
    ".xbarrier": None,
}

_OBJECT_SIZES = {
    "/c/void-pointer": {
        "commands": ["sizeof(void *)"],
        "uid": "/rtems/val/mem-basic"
    },
    "/rtems/barrier/info": {
        "commands": ["sizeof(_Barrier_Information)"],
        "uid": "/rtems/barrier/val/mem-wait-rel-del"
    },
    "/rtems/barrier/obj": {
        "commands": ["sizeof(Objects_Control *) + sizeof(Barrier_Control)"],
        "uid": "/rtems/barrier/val/mem-wait-rel-del"
    },
    "/rtems/message/info": {
        "commands": ["sizeof(_Message_queue_Information)"],
        "uid": "/rtems/message/val/mem-snd-rcv-del"
    },
    "/rtems/message/obj": {
        "commands":
        ["sizeof(Objects_Control *) + sizeof(Message_queue_Control)"],
        "uid": "/rtems/message/val/mem-snd-rcv-del"
    },
    "/rtems/part/info": {
        "commands": ["sizeof(_Partition_Information)"],
        "uid": "/rtems/part/val/mem-get-ret-del"
    },
    "/rtems/part/obj": {
        "commands": ["sizeof(Objects_Control *) + sizeof(Partition_Control)"],
        "uid": "/rtems/part/val/mem-get-ret-del"
    },
    "/rtems/priority/control": {
        "commands": ["sizeof(Priority_Control)"],
        "uid": "/rtems/sem/val/mem-obt-rel-del"
    },
    "/rtems/ratemon/info": {
        "commands": ["sizeof(_Rate_monotonic_Information)"],
        "uid": "/rtems/ratemon/val/mem-period-del"
    },
    "/rtems/ratemon/obj": {
        "commands":
        ["sizeof(Objects_Control *) + sizeof(Rate_monotonic_Control)"],
        "uid": "/rtems/ratemon/val/mem-period-del"
    },
    "/rtems/sem/info": {
        "commands": ["sizeof(_Semaphore_Information)"],
        "uid": "/rtems/sem/val/mem-obt-rel-del"
    },
    "/rtems/sem/obj": {
        "commands": ["sizeof(Objects_Control *) + sizeof(Semaphore_Control)"],
        "uid": "/rtems/sem/val/mem-obt-rel-del"
    },
    "/rtems/sem/obj-mrsp": {
        "commands": [
            "sizeof(Objects_Control *) + sizeof(Objects_Control) + "
            "sizeof(MRSP_Control)"
        ],
        "uid":
        "/rtems/sem/val/mem-obt-rel-del"
    },
    "/rtems/task/info": {
        "commands": ["sizeof(_RTEMS_tasks_Information)"],
        "uid": "/rtems/val/mem-basic"
    },
    "/rtems/task/obj": {
        "commands": [
            "sizeof(Objects_Control *) + sizeof(Thread_Control) + "
            "sizeof(RTEMS_API_Control) + sizeof(Thread_queue_Heads)",
            "sizeof(Objects_Control *) + sizeof(Thread_Control) + "
            "sizeof(RTEMS_API_Control)"
        ],
        "uid":
        "/rtems/val/mem-basic"
    },
    "/rtems/timer/info": {
        "commands": ["sizeof(_Timer_Information)"],
        "uid": "/rtems/timer/val/mem-after"
    },
    "/rtems/timer/obj": {
        "commands": ["sizeof(Objects_Control *) + sizeof(Timer_Control)"],
        "uid": "/rtems/timer/val/mem-after"
    },
    "/rtems/userext/info": {
        "commands": ["sizeof(_Extension_Information)"],
        "uid": "/rtems/userext/val/mem-create"
    },
    "/rtems/userext/obj": {
        "commands": ["sizeof(Objects_Control *) + sizeof(Extension_Control)"],
        "uid": "/rtems/userext/val/mem-create"
    },
    "/scheduler/control": {
        "commands": ["sizeof(Scheduler_Control)"],
        "uid": "/rtems/val/mem-basic"
    },
    "/scheduler/priority/context": {
        "commands": ["sizeof(Scheduler_priority_Context)"],
        "uid": "/rtems/val/mem-basic"
    },
    "/scheduler/priority/node": {
        "commands": ["sizeof(Scheduler_priority_Node)"],
        "uid": "/rtems/val/mem-basic"
    },
    "/scheduler/smp/edf/context": {
        "commands": [
            "sizeof(Scheduler_EDF_SMP_Context) + "
            "sizeof(Scheduler_EDF_SMP_Ready_queue)"
        ],
        "uid":
        "/rtems/val/mem-smp-1"
    },
    "/scheduler/smp/edf/node": {
        "commands": ["sizeof(Scheduler_EDF_SMP_Node)"],
        "uid": "/rtems/val/mem-smp-1"
    },
    "/scheduler/smp/edf/ready-queue": {
        "commands": ["sizeof(Scheduler_EDF_SMP_Ready_queue)"],
        "uid": "/rtems/val/mem-smp-1"
    },
    "/score/chain/control": {
        "commands": ["sizeof(Chain_Control)"],
        "uid": "/rtems/val/mem-basic"
    },
    "/score/tq/priority-queue": {
        "commands": ["sizeof(Thread_queue_Priority_queue)"],
        "uid": "/rtems/val/mem-basic"
    },
}


def _section_key(key_value: tuple[str, Any]) -> int:
    return _SECTIONS[key_value[0]]


def _get_sections_of_item(sections_by_uid: SectionsByUID,
                          item: Item) -> tuple[int, ...]:
    return tuple(size for _, size in sorted(
        sections_by_uid.get(item.uid, {}).items(), key=_section_key))


def gather_sections(item_cache: ItemCache, path: str, objdump: str,
                    gdb: str) -> dict[str, dict[str, int]]:
    """
    Gathers the object sizes for all memory benchmarks of the item cache using
    the programs of the path.
    """
    sections_by_uid: dict[str, dict[str, int]] = {}
    for item in item_cache.items_by_type["memory-benchmark"]:
        sections = _get_sections(item, path, objdump, gdb)
        if sections:
            sections_by_uid[item.uid] = sections
    return sections_by_uid


def gather_object_sizes(item_cache: ItemCache, path: str,
                        gdb: str) -> dict[str, int]:
    """
    Gathers the object sizes for all memory benchmarks of the item cache using
    the programs of the path.
    """
    object_sizes: dict[str, int] = {}
    for key, value in _OBJECT_SIZES.items():
        try:
            item = item_cache[value["uid"]]
        except KeyError:
            continue
        file = _get_path_to_test_suite_elf_file(item, path)
        for command in value["commands"]:
            cmd = [gdb, "-ex", f"p {command}", "--batch", ""]
            cmd[-1] = file
            stdout: list[str] = []
            status = run_command(cmd, stdout=stdout)
            if status == 0:
                output = " ".join(stdout)
                match = re.search(r"\$[0-9]+\s*=\s*([0-9]+)", output)
                assert match
                object_sizes[key] = int(match.group(1))
                break
    return object_sizes


def _do_gather_test_suites(items: list[Item], item: Item) -> None:
    if item.type == "memory-benchmark":
        items.append(item)
    # Gather the validation items first so that the root validation item shows
    # up first.
    for child in item.children("validation"):
        _do_gather_test_suites(items, child)
    for child in item.children("requirement-refinement"):
        _do_gather_test_suites(items, child)


def _gather_benchmarks(root: Item) -> list[Item]:
    """ Gather all test suite items related to the root item. """
    items: list[Item] = []
    _do_gather_test_suites(items, root)
    return items


def _get_path_to_test_suite_elf_file(item: Item, path: str) -> str:
    """ Returns the path to the ELF file of the test suite. """
    name = item["test-target"]
    return os.path.join(path, f"{name[:name.rfind('.')]}.norun.exe")


def _try_add_workspace(gdb: str, elf: str,
                       section_limits: dict[str, tuple[int, int]]) -> None:
    if ".noinit" in section_limits:
        return
    # Maybe an older version of RTEMS
    stdout: list[str] = []
    cmd = [gdb, "-ex", "p Configuration.work_space_size", "--batch", elf]
    status = run_command(cmd, stdout=stdout)
    if status == 0 and stdout[0].startswith("$1 = "):
        section_limits[".noinit"] = (0, int(stdout[0][5:]))


def _get_size(section_limits: dict[str, tuple[int, int]], key: str) -> int:
    limits = section_limits.get(key, (0, 0))
    return limits[1] - limits[0]


def _run_objdump(objdump: str, elf: str) -> list[str]:
    stdout: list[str] = []
    status = run_command([objdump, "-h", elf], stdout=stdout)
    if status != 0:
        return []
    return stdout


def _make_sections(
        section_limits: dict[str, tuple[int, int]]) -> dict[str, int]:
    sections: dict[str, int] = {}
    for key in _SECTIONS:
        sections[key] = _get_size(section_limits, key)
    sections[".noinit"] += _get_size(section_limits, ".vector")
    return sections


def _get_sections(item: Item, path_to_elf_files: str, objdump: str,
                  gdb: str) -> dict[str, int]:
    """
    Get the memory benchmark sections of the program associated with the item
    in the path to ELF files.
    """
    elf = _get_path_to_test_suite_elf_file(item, path_to_elf_files)
    stdout = _run_objdump(objdump, elf)
    if not stdout:
        return {}
    section_limits: dict[str, tuple[int, int]] = {}
    for line in stdout:
        match = _SECTION.search(line)
        if match:
            name = match.group(1)
            size = int(match.group(2), 16)
            section = _SECTION_MAP[name]
            if size != 0 and section:
                start = int(match.group(3), 16)
                end = start + size
                limits = section_limits.get(section, (2**64, 0))
                section_limits[section] = (min(limits[0],
                                               start), max(limits[1], end))
    _try_add_workspace(gdb, elf, section_limits)
    return _make_sections(section_limits)


def _make_label(scope: str, item: Item) -> str:
    return f"{scope}BenchmarkSpec{item.ident}"


def _generate_table(content: SphinxContent, sections_by_uid: SectionsByUID,
                    items: list[Item]) -> None:
    scope = content.label
    rows: list[tuple[str, ...]] = []
    for index, item in enumerate(items):
        sections = _get_sections_of_item(sections_by_uid, item)
        spec = (get_reference(_make_label(scope, item), item.uid), )
        if index == 0:
            assert sections
            keys = ("Specification", ) + _SECTION_KEYS
            base = sections
            rows.append(keys)
            rows.append(spec + tuple(map(str, sections)))
        elif sections:
            rows.append(spec + tuple(f"{i - j:+}"
                                     for i, j in zip(sections, base)))
        else:
            rows.append(spec + tuple("?") * len(keys))

    pivot = items[0]
    with content.section(f"Benchmarks based on: {pivot.spec}"):
        content.add(f"""The following static memory benchmarks are based on the
reference memory benchmark specified by
{get_reference(_make_label(scope, pivot), pivot.spec_2)}.
The numbers of the first row represent the section sizes of the reference
memory benchmark program in bytes.  The numbers in the following rows indicate
the change in bytes of the section sizes with respect to the reference memory
benchmark program of the first row.  A ``+`` indicates a size increase and a
``-`` indicates a size decrease.  This hints how the static memory usage
changes when the feature set changes with respect to the reference memory
benchmark.""")
        content.add_simple_table(rows,
                                 widths=[50, 10, 10, 10, 10, 10],
                                 font_size=-3)


_WARNING_NO_MEMBENCH = """.. topic:: WARNING

    There are no results available for this static memory usage benchmark.
"""


def _generate_paragraphs(content: SphinxContent,
                         sections_by_uid: SectionsByUID, mapper: ItemMapper,
                         items: list[Item]) -> None:
    for item in items:
        with content.section(f"Benchmark: {item.spec}"):
            content.wrap(mapper.substitute(item["test-brief"], item))
            content.wrap(mapper.substitute(item["test-description"], item))
            sections = _get_sections_of_item(sections_by_uid, item)
            if sections:
                content.add_simple_table(
                    [_SECTION_KEYS, tuple(map(str, sections))],
                    widths=[20, 20, 20, 20, 20],
                    font_size=-3)
            else:
                content.add(_WARNING_NO_MEMBENCH)


def _generate_tables(content: SphinxContent, sections_by_uid: SectionsByUID,
                     root: Item, table_pivots: list[str]) -> list[Item]:
    root_items = _gather_benchmarks(root)
    _generate_table(content, sections_by_uid, root_items)
    for pivot_uid in table_pivots:
        pivot = root.map(pivot_uid)
        if not pivot.enabled:
            continue
        pivot_items = _gather_benchmarks(pivot)
        _generate_table(content, sections_by_uid, pivot_items)
    return root_items


def generate_tables(content: SphinxContent, sections_by_uid: SectionsByUID,
                    root: Item, table_pivots: list[str]) -> None:
    """ Generate memory benchmark tables. """
    _generate_tables(content, sections_by_uid, root, table_pivots)


def generate_variants_table(content: SphinxContent,
                            sections_by_build_label: dict[str,
                                                          dict[str,
                                                               SectionsByUID]],
                            root: Item,
                            variants: list[MembenchVariant]) -> None:
    """ Generate memory benchmark variant comparison tables. """
    items = _gather_benchmarks(root)
    rows: list[tuple[str | int,
                     ...]] = [("Specification", "Variant") + _SECTION_KEYS]
    for item in items:
        spec: str | int = get_reference(_make_label(content.label, item),
                                        item.uid)
        for index, variant in enumerate(variants):
            sections = _get_sections_of_item(
                sections_by_build_label[variant.build_label]["membench"], item)
            what = (spec, variant.name)
            if index == 0:
                assert sections
                base = sections
                rows.append(what + tuple(map(str, sections)))
            else:
                if sections:
                    rows.append(what + tuple(f"{i - j:+}"
                                             for i, j in zip(sections, base)))
                else:
                    rows.append(what + tuple("?") * len(base))
            spec = COL_SPAN
    content.add_grid_table(rows, [35, 20, 9, 9, 9, 9, 9], font_size=-3)


def generate(content: SphinxContent, sections_by_uid: SectionsByUID,
             root: Item, table_pivots: list[str], mapper: ItemMapper) -> None:
    """
    Generate memory benchmark documentation for items dependent on the root
    item and executables in the path.
    """
    root_items = _generate_tables(content, sections_by_uid, root, table_pivots)
    _generate_paragraphs(content, sections_by_uid, mapper, root_items)



# SPDX-License-Identifier: BSD-2-Clause
""" Run actions. """

# Copyright (C) 2022, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import copy
import contextlib
import dataclasses
import itertools
import os
import logging
from pathlib import Path
import shutil
from typing import Any, Callable, Iterator, Optional, TextIO

from specitems import (Item, ItemGetValueContext, IS_ENABLED_OPS,
                       escape_code_line, is_enabled_with_ops,
                       run_subprocess_action)

from .dirstatebase import DirectoryStateBase
from .pkgitems import BuildItem, BuildItemMapper, PackageBuildDirector
from .util import copy_file, copy_and_substitute, remove_empty_directories


def _get_host_processor_count(_ctx: ItemGetValueContext) -> str:
    count = os.cpu_count()
    return str(count if count is not None else 1)


@dataclasses.dataclass
class _Env:
    clear: list[str] = dataclasses.field(default_factory=list)
    path: dict[str, tuple[list[str],
                          list[str]]] = dataclasses.field(default_factory=dict)
    var_set: dict[str, str] = dataclasses.field(default_factory=dict)
    var_unset: list[str] = dataclasses.field(default_factory=list)


def _env_clear(env: _Env, _action: dict[str, str]) -> None:
    env.clear = ["-i"]


def _env_ignore(_env: _Env, _action: dict[str, str]) -> None:
    pass


def _env_path_append(env: _Env, action: dict[str, str]) -> None:
    env.path.setdefault(action["name"], ([], []))[1].append(action["value"])


def _env_path_prepend(env: _Env, action: dict[str, str]) -> None:
    env.path.setdefault(action["name"], ([], []))[0].append(action["value"])


def _env_set(env: _Env, action: dict[str, str]) -> None:
    name = action["name"]
    env.var_set[name] = action["value"]
    env.path.pop(name, None)


def _env_unset(env: _Env, action: dict[str, str]) -> None:
    name = action["name"]
    env.var_unset.append(name)
    env.path.pop(name, None)
    env.var_set.pop(name, None)


_ENV_ACTIONS = {
    "clear": _env_clear,
    "ignore": _env_ignore,
    "path-append": _env_path_append,
    "path-prepend": _env_path_prepend,
    "set": _env_set,
    "unset": _env_unset
}


def _add_description(client: BuildItem, line: str) -> None:
    client.description.append(escape_code_line(line))


def _copy_and_op(client: BuildItem, action: dict,
                 output: Optional[DirectoryStateBase],
                 operation: Callable[[BuildItem, str, str], None]) -> None:
    action = client.substitute(action)
    assert isinstance(output, DirectoryStateBase)
    input_name = action["input-name"]
    if input_name == ".":
        input_state = client
    else:
        input_state = client.input(input_name)
    assert isinstance(input_state, DirectoryStateBase)
    source = action["source"]
    source_base = input_state.directory
    target_base = output.directory
    if source is None:
        prefix = action["target"]
        if prefix is None:
            prefix = "."
        targets: list[str] = []
        for source_file in input_state:
            tail = os.path.relpath(source_file, source_base)
            target_file = os.path.join(target_base, prefix, tail)
            targets.append(tail)
            _add_description(client, f"cp {source_file} {target_file}")
            operation(client, source_file, target_file)
        output.add_files(targets)
    else:
        source_file = os.path.join(source_base, source)
        target = action["target"]
        if target is None:
            target_file = output.file
        else:
            output.add_files([target])
            target_file = os.path.join(target_base, target)
        _add_description(client, f"cp {source_file} {target_file}")
        operation(client, source_file, target_file)


def _do_copy(client: BuildItem, source_file: str, target_file: str) -> None:
    copy_file(source_file, target_file, client.uid)


def _do_copy_and_substitute(client: BuildItem, source_file: str,
                            target_file: str) -> None:
    copy_and_substitute(source_file, target_file, client.mapper, client.uid)


class RunActionsProvider:
    """ Runs actions provider. """

    # pylint: disable=too-few-public-methods

    def __init__(self, client: BuildItem) -> None:
        self._client = client
        client.mapper.add_get_value(
            f"{client.item.type}:/host-processor-count",
            _get_host_processor_count)
        self._is_enabled_ops = IS_ENABLED_OPS | {
            "eq": self._equal,
            "has-output": self._has_output
        }

    def _run_actions(self, client: BuildItem, actions: dict) -> None:
        for index, action in enumerate(actions):
            action_type = action["action"]
            logging.info("%s: run action %i: %s", client.uid, index,
                         action_type)
            if is_enabled_with_ops(client.enabled_set,
                                   client.substitute(action["enabled-by"]),
                                   self._is_enabled_ops):
                output_name = action.get("output-name", None)
                if output_name is None:
                    output: None | BuildItem = None
                elif output_name == ".":
                    output = client
                else:
                    try:
                        output = client.output(output_name)
                    except KeyError:
                        continue
                assert output is None or isinstance(output, DirectoryStateBase)
                RunActionsProvider._ACTIONS[action_type](self, client, action,
                                                         output)

    def run(self, actions: dict) -> None:
        """ Run the actions. """
        if not actions:
            return
        client = self._client
        client.description.add("Run the following actions:")
        with client.description.directive("code-block", "none", [":linenos:"]):
            self._run_actions(client, actions)

    def _equal(self, _enabled_set: list[str], enabled_by: Any,
               _ops: dict) -> bool:
        first = enabled_by[0]
        for value in enabled_by[1:]:
            if first != value:
                return False
        return True

    def _has_output(self, _enabled_set: list[str], enabled_by: Any,
                    _ops: dict) -> bool:
        assert isinstance(enabled_by, str)
        try:
            self._client.output(enabled_by)
        except KeyError:
            return False
        return True

    def _copy(self, client: BuildItem, action: dict,
              output: Optional[DirectoryStateBase]) -> None:
        _copy_and_op(client, action, output, _do_copy)

    def _copy_and_substitute(self, client: BuildItem, action: dict,
                             output: Optional[DirectoryStateBase]) -> None:
        _add_description(client, "# substitute content")
        _copy_and_op(client, action, output, _do_copy_and_substitute)

    @contextlib.contextmanager
    def _create_target(
            self, client: BuildItem, action: dict,
            output: Optional[DirectoryStateBase]
    ) -> Iterator[tuple[str, TextIO]]:
        assert isinstance(output, DirectoryStateBase)
        target = client.substitute(action["target"])
        if target is None:
            target = output.file
        else:
            output.add_files([target])
            target = os.path.join(output.directory, target)
        logging.info("%s: create: %s", client.uid, target)
        os.makedirs(os.path.dirname(target), exist_ok=True)
        yield target, open(target, "w", encoding="utf-8")

    def _is_enabled(self, component: dict) -> bool:
        client = self._client
        return is_enabled_with_ops(client.enabled_set,
                                   client.substitute(component["enabled-by"]),
                                   self._is_enabled_ops)

    def _create_ini_file(self, client: BuildItem, action: dict,
                         output: Optional[DirectoryStateBase]) -> None:
        with self._create_target(client, action, output) as (target, file):
            for section in action["sections"]:
                if not self._is_enabled(section):
                    continue
                name = client.substitute(section["name"])
                file.write(f"[{name}]\n")
                for key_value in section["key-value-pairs"]:
                    if not self._is_enabled(key_value):
                        continue
                    key_value = client.substitute(key_value)
                    file.write(f"{key_value['key']} = {key_value['value']}\n")
            _add_description(client, f"# create INI file {target}")

    def _create_kconfig_file(self, client: BuildItem, action: dict,
                             output: Optional[DirectoryStateBase]) -> None:
        with self._create_target(client, action, output) as (target, file):
            for key_value in action["key-value-pairs"]:
                if self._is_enabled(key_value):
                    key_value = client.substitute(key_value)
                    file.write(f"{key_value['key']}={key_value['value']}\n")
            _add_description(client, f"# create Kconfig file {target}")

    def _directory_state_clear(self, client: BuildItem, _action: dict,
                               output: Optional[DirectoryStateBase]) -> None:
        assert isinstance(output, DirectoryStateBase)
        output.clear()
        _add_description(client, f"# clear directory state {output.item.spec}")

    def _directory_state_discard(self, client: BuildItem, _action: dict,
                                 output: Optional[DirectoryStateBase]) -> None:
        assert isinstance(output, DirectoryStateBase)
        output.discard()
        _add_description(client,
                         f"# discard directory state {output.item.spec}")

    def _directory_state_discard_and_clear(
            self, client: BuildItem, action: dict,
            output: Optional[DirectoryStateBase]) -> None:
        self._directory_state_discard(client, action, output)
        self._directory_state_clear(client, action, output)

    def _directory_state_discard_excludes(
            self, client: BuildItem, _action: dict,
            output: Optional[DirectoryStateBase]) -> None:
        assert isinstance(output, DirectoryStateBase)
        for exclude_item in client.item.parents("directory-state-exclude"):
            exclude_state = client.director[exclude_item.uid]
            assert isinstance(exclude_state, DirectoryStateBase)
            exclude_state.discard()
            _add_description(
                client, f"# discard directory state {exclude_state.item.spec}")

    def _directory_state_add_files(
            self, client: BuildItem, action: dict,
            output: Optional[DirectoryStateBase]) -> None:
        action = client.substitute(action)
        assert isinstance(output, DirectoryStateBase)
        root = Path(action["path"]).absolute()
        pattern = action["pattern"]
        logging.info("%s: add files matching '%s' in: %s", client.uid, pattern,
                     root)
        base = output.directory
        output.add_files([
            os.path.relpath(path, base) for path in root.glob(pattern)
            if path.is_symlink() or not path.is_dir()
        ])
        _add_description(client,
                         f"# add files to directory state {output.item.spec}")

    def _directory_state_add_tarfile_members(
            self, client: BuildItem, action: dict,
            output: Optional[DirectoryStateBase]) -> None:
        action = client.substitute(action)
        assert isinstance(output, DirectoryStateBase)
        root = Path(action["search-path"])
        pattern = action["pattern"]
        logging.info("%s: search for tarfiles matching '%s' in: %s",
                     client.uid, pattern, root)
        for path in root.glob(pattern):
            output.add_tarfile_members(path, action["prefix-path"],
                                       action["extract"])
            _add_description(
                client, f"# add files from {path} to "
                f"directory state {output.item.spec}")

    def _directory_state_tree_op(self, client: BuildItem, action: dict,
                                 output: DirectoryStateBase,
                                 tree_op: Any) -> tuple[str, str]:
        action = client.substitute(action)
        root = Path(action["root"]).absolute()
        prefix = action["prefix"]
        if prefix is None:
            prefix = "."
        tree_op(output, root, prefix, action["excludes"])
        return os.path.normpath(root), os.path.normpath(
            os.path.join(output.directory, prefix))

    def _directory_state_add_tree(
            self, client: BuildItem, action: dict,
            output: Optional[DirectoryStateBase]) -> None:
        assert isinstance(output, DirectoryStateBase)
        source, _ = self._directory_state_tree_op(client, action, output,
                                                  DirectoryStateBase.add_tree)
        _add_description(
            client, f"# add files of directory {source} "
            f"to directory state {output.item.spec}")

    def _directory_state_copy_tree(
            self, client: BuildItem, action: dict,
            output: Optional[DirectoryStateBase]) -> None:
        assert isinstance(output, DirectoryStateBase)
        source, target = self._directory_state_tree_op(
            client, action, output, DirectoryStateBase.copy_tree)
        _add_description(client, f"mkdir -p {target}")
        _add_description(
            client,
            f"# add files of below copy to directory state {output.item.spec}")
        _add_description(client, f"cp -r {source} {target}")

    def _directory_state_move_tree(
            self, client: BuildItem, action: dict,
            output: Optional[DirectoryStateBase]) -> None:
        assert isinstance(output, DirectoryStateBase)
        source, target = self._directory_state_tree_op(
            client, action, output, DirectoryStateBase.move_tree)
        _add_description(
            client,
            f"# add files of below move to directory state {output.item.spec}")
        _add_description(
            client, f"rsync -a --remove-source-files {source}/* {target}")

    def _directory_state_copy_file(
        self,
        client: BuildItem,
        action: dict,
        output: Optional[DirectoryStateBase],
    ) -> None:
        action = client.substitute(action)
        assert isinstance(output, DirectoryStateBase)
        source_file = action["source"]
        input_name = action.get("input-name", None)
        if input_name is not None:
            if input_name == ".":
                input_state = client
            else:
                input_state = client.input(input_name)
            assert isinstance(input_state, DirectoryStateBase)
            source_file = os.path.join(input_state.directory, source_file)
        target = action["target"]
        output.copy_file(source_file, target)
        _add_description(
            client,
            f"# add files of below copy to directory state {output.item.spec}")
        _add_description(
            client,
            f"cp {source_file} {os.path.join(output.directory, target)}")

    def _process(self, client: BuildItem, action: dict,
                 _output: Optional[DirectoryStateBase]) -> None:
        action = client.substitute(action)
        run_subprocess_action(client.uid, action)
        cwd = action["working-directory"]
        if cwd != ".":
            cmd = f"cd {cwd} && "
        else:
            cmd = ""
        env_actions = action["env"]
        if env_actions:
            env = _Env()
            for env_action in env_actions:
                _ENV_ACTIONS[env_action["action"]](env, env_action)
            args = []
            args.extend(env.clear)
            args.extend(f"-u {name}" for name in env.var_unset)
            for name, values in sorted(env.path.items()):
                value = ":".join(
                    itertools.chain(values[0],
                                    [env.var_set.get(name, f"${{{name}}}")],
                                    values[1]))
                args.append(f"{name}={value}")
                env.var_set.pop(name, None)
            args.extend(f"{name}={value}"
                        for name, value in sorted(env.var_set.items()))
            cmd = f"{cmd}env {' '.join(args)} "
        _add_description(client, f"{cmd}{' '.join(action['command'])}")

    def _for_each(self, client: BuildItem, action: dict,
                  _output: Optional[DirectoryStateBase]) -> None:
        for link, _ in client.input_links("run-actions-member"):
            params = client.item["params"]
            client.item["params"] = copy.deepcopy(params)
            client.item["params"].update(link["params"])
            self._run_actions(client, action["member-actions"])
            client.item["params"] = params

    def _mkdir(self, client: BuildItem, action: dict,
               _output: Optional[DirectoryStateBase]) -> None:
        action = client.substitute(action)
        path = Path(action["path"])
        logging.info("%s: make directory: %s", client.uid, path)
        path.mkdir(parents=action["parents"], exist_ok=action["exist-ok"])
        _add_description(client, f"mkdir -p {path}")

    def _remove_path(self, client: BuildItem, path: Path) -> None:
        if path.is_dir():
            logging.info("%s: remove directory: %s", client.uid, path)
            path.rmdir()
            _add_description(client, f"rm -r {path}")
        else:
            logging.info("%s: unlink file: %s", client.uid, path)
            path.unlink()
            _add_description(client, f"rm {path}")

    def _remove(self, client: BuildItem, action: dict,
                _output: Optional[DirectoryStateBase]) -> None:
        action = client.substitute(action)
        path = Path(action["path"])
        if action["missing-ok"]:
            try:
                self._remove_path(client, path)
            except FileNotFoundError:
                pass
        else:
            self._remove_path(client, path)

    def _remove_empty_directories(
            self, client: BuildItem, action: dict,
            _output: Optional[DirectoryStateBase]) -> None:
        path = client.substitute(action["path"])
        remove_empty_directories(client.uid, path)
        _add_description(client,
                         f"find {path} -type d -empty -exec rmdir {{}} \\;")

    def _remove_glob(self, client: BuildItem, action: dict,
                     _output: Optional[DirectoryStateBase]) -> None:
        action = client.substitute(action)
        root = Path(action["path"])
        for pattern in action["patterns"]:
            logging.info(
                "%s: remove files and directories matching with '%s' in: %s",
                client.uid, pattern, root)
            for path in root.glob(pattern):
                if path.is_symlink() or not path.is_dir():
                    logging.info("%s: remove file: %s", client.uid, path)
                    path.unlink()
                    _add_description(client, f"rm {path}")
                elif action["remove-tree"]:
                    logging.info("%s: remove directory tree: %s", client.uid,
                                 path)
                    shutil.rmtree(path)
                    _add_description(client, f"rm -r {path}")
                else:
                    logging.info("%s: remove directory: %s", client.uid, path)
                    path.rmdir()
                    _add_description(client, f"rm -r {path}")

    def _remove_tree(self, client: BuildItem, action: dict,
                     _output: Optional[DirectoryStateBase]) -> None:
        action = client.substitute(action)
        path = action["path"]
        logging.info("%s: remove directory tree: %s", client.uid, path)
        if action["missing-ok"]:
            try:
                shutil.rmtree(path)
            except FileNotFoundError:
                pass
            _add_description(client, f"rm -rf {path}")
        else:
            shutil.rmtree(path)
            _add_description(client, f"rm -r {path}")

    def _touch(self, client: BuildItem, action: dict,
               _output: Optional[DirectoryStateBase]) -> None:
        action = client.substitute(action)
        path = Path(action["path"])
        logging.info("%s: touch file: %s", client.uid, path)
        path.touch(exist_ok=action["exist-ok"])
        _add_description(client, f"touch {path}")

    _ACTIONS = {
        "copy": _copy,
        "copy-and-substitute": _copy_and_substitute,
        "create-ini-file": _create_ini_file,
        "create-kconfig-file": _create_kconfig_file,
        "directory-state-add-files": _directory_state_add_files,
        "directory-state-add-tarfile-members":
        _directory_state_add_tarfile_members,
        "directory-state-add-tree": _directory_state_add_tree,
        "directory-state-clear": _directory_state_clear,
        "directory-state-copy-file": _directory_state_copy_file,
        "directory-state-copy-tree": _directory_state_copy_tree,
        "directory-state-discard": _directory_state_discard,
        "directory-state-discard-and-clear":
        _directory_state_discard_and_clear,
        "directory-state-discard-excludes": _directory_state_discard_excludes,
        "directory-state-move-tree": _directory_state_move_tree,
        "for-each": _for_each,
        "mkdir": _mkdir,
        "remove": _remove,
        "remove-empty-directories": _remove_empty_directories,
        "remove-glob": _remove_glob,
        "remove-tree": _remove_tree,
        "subprocess": _process,
        "touch": _touch
    }


class RunActions(BuildItem):
    """ Runs actions. """

    def __init__(self,
                 director: PackageBuildDirector,
                 item: Item,
                 mapper: Optional[BuildItemMapper] = None) -> None:
        super().__init__(director, item, mapper)
        self._run_actions_provider = RunActionsProvider(self)

    def run(self) -> None:
        self._run_actions_provider.run(self.item["actions"])



# SPDX-License-Identifier: BSD-2-Clause
""" Provides the basic support to build a package. """

# Copyright (C) 2021 EDISOFT
# Copyright (C) 2020, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from contextlib import contextmanager
import copy
import fnmatch
import graphlib
import itertools
import os
import logging
import subprocess
from typing import Any, Callable, Iterable, Iterator, Optional, Type, Union

from specitems import (augment_glossary_terms, data_digest, EnabledSet,
                       is_enabled, Item, ItemCache, ItemDataByUID,
                       ItemGetValueContext, ItemGetValue, ItemMapper,
                       ItemSelection, ItemView, Link, link_is_enabled,
                       pickle_load_data_by_uid, SphinxContent, SphinxMapper,
                       to_iterable)
from specware import SpecWareTypeProvider, run_command


class BuildItemTypeProvider(SpecWareTypeProvider):
    """ Provides a type system for specification build items. """

    def __init__(self, data_by_uid: ItemDataByUID) -> None:
        data_by_uid.update(
            pickle_load_data_by_uid(
                os.path.join(os.path.dirname(__file__), "spec.pickle")))
        super().__init__(data_by_uid)
        dict_info = self.data_by_uid["/spec/root"]["spec-info"]["dict"]
        dict_info["mandatory-attributes"] = list(
            sorted(dict_info["attributes"].keys()))
        dict_info["attributes"]["workspace-hash"] = {
            "description": None,
            "spec-type": "sha256"
        }


def _git_has_staged_files(directory: str) -> bool:
    stdout: list[str] = []
    status = run_command(["git", "diff", "--name-only", "--cached"], directory,
                         stdout)
    assert status == 0
    return bool(stdout)


def _get_input_links(item: Item) -> Iterator[Link]:
    yield from itertools.chain(item.links_to_parents("input"),
                               item.links_to_children("input-to"))


def build_item_input(item: Item, name: str) -> Item:
    """ Return the first input item with the name.  """
    for link in _get_input_links(item):
        if link["name"] == name:
            return link.item
    raise KeyError


def _get_spec(ctx: ItemGetValueContext) -> Any:
    return ctx.item.spec


def _basename(value: str) -> str:
    return os.path.basename(value)


def _dash(value: str) -> str:
    return f"-{value}" if value else ""


def _dirname(value: str) -> str:
    return os.path.dirname(value)


def _relpath(value: str, start: str) -> str:
    return os.path.relpath(value, start)


def _slash(value: str) -> str:
    return f"/{value}" if value else ""


class BuildItemMapper(SphinxMapper):
    """
    The build item mapper provides a method to get a link to the primary
    documentation place of the item.
    """

    def __init__(self, item: Item):
        super().__init__(item)
        self.add_default_get_value("spec", _get_spec)
        self.add_value_transformer("basename", _basename)
        self.add_value_transformer("dirname", _dirname)
        self.add_value_transformer("dash", _dash)
        self.add_value_transformer("relpath", _relpath)
        self.add_value_transformer("slash", _slash)

    def get_link(self, item: Item, document_key: None | str = None) -> str:
        """
        Return a link to the item.

        The document key selects the preferred target document.  If it is not
        available for item, then the default document will be the link target.
        """
        raise NotImplementedError


def _get_input_or_output(ctx: ItemGetValueContext,
                         what: Callable[[str], "BuildItem"]) -> Any:
    name_end = ctx.remaining_path.find("/")
    build_item = what(ctx.remaining_path[:name_end])
    return ctx.reset(build_item.item, ctx.remaining_path[name_end + 1:])


class BuildItem():
    """ Represents a package build item. """

    # pylint: disable=too-many-public-methods
    @classmethod
    def prepare_factory(cls, _factory: "BuildItemFactory",
                        _type_name: str) -> None:
        """ Prepare the build item factory for the type. """

    def __init__(self,
                 director: "PackageBuildDirector",
                 item: Item,
                 mapper: Optional[BuildItemMapper] = None) -> None:
        self.director = director
        self.item = item
        if mapper is None:
            mapper = BuildItemMapper(item)
        self.mapper = mapper
        try:
            component = self.input("component")
            assert isinstance(component, PackageComponent)
        except KeyError:
            component = None
        self._component_stack = [component]
        mapper.add_default_get_value("component", self._get_component)
        director.factory.add_get_values_to_mapper(self.mapper)
        my_type = self.item.type
        self.mapper.add_get_value(f"{my_type}:/input", self._get_input)
        self.mapper.add_get_value(f"{my_type}:/output", self._get_output)
        self.description = SphinxContent()
        self.description.add(item.get("build-description", None))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BuildItem):
            return NotImplemented
        return self.item == other.item  # pylint: disable=protected-access

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, BuildItem):
            return NotImplemented
        return self.item < other.item  # pylint: disable=protected-access

    def __hash__(self) -> int:
        return hash(self.item)

    def __contains__(self, key: str) -> bool:
        return key in self.item

    def __getitem__(self, key: str) -> Any:
        return self.substitute(self.item[key])

    def __setitem__(self, key: str, value: Any) -> None:
        self.item[key] = value

    def get(self, key: str, default: Any) -> Any:
        """
        Get the substitution of the attribute value if the attribute exists,
        otherwise the substitution of the default value is returned.
        """
        return self.substitute(self.item.get(key, default))

    @property
    def uid(self) -> str:
        """ Is the UID of the build item. """
        return self.item.uid

    @property
    def enabled_set(self) -> EnabledSet:
        """ Is the enabled set of the build item. """
        return self.item.cache.enabled_set

    @property
    def enabled(self) -> bool:
        """ Is true, if the build item is enabled, otherwise false.  """
        return self.item.enabled

    @property
    def component(self) -> "PackageComponent":
        """ Is the package component of the build item. """
        the_component = self._component_stack[-1]
        if the_component is None:
            raise StopIteration
        return the_component

    @component.setter
    def component(self, the_component: "PackageComponent") -> None:
        """ Set the package component of the build item. """
        self._component_stack[-1] = the_component

    @property
    def component_depth(self) -> int:
        """ Is the package component stack depth of the build item. """
        return len(self._component_stack)

    @property
    def git_directory(self) -> str:
        """ Is the package Git directory. """
        return self.director.package["deployment-directory"]

    def push_component(self, component: "PackageComponent") -> None:
        """
        Push the component to the top of the package component stack.
        """
        self._component_stack.append(component)

    def pop_component(self) -> None:
        """ Pop the top from the package component stack. """
        self._component_stack.pop()

    @contextmanager
    def component_scope(self, component: "PackageComponent") -> Iterator[None]:
        """ Opens a package component context. """
        self.push_component(component)
        item_cache = self.item.cache
        item_cache.push_selection(component.selection)
        item_cache.push_view(component.view)
        yield
        item_cache.pop_view()
        item_cache.pop_selection()
        self.pop_component()

    def do_run(self) -> None:
        """ Run the build. """
        self.mapper.item = self.item
        self.discard_outputs()
        self.clear_description()
        logging.info("%s: run", self.uid)
        self.run()
        self.refresh_outputs()
        self.refresh_make_and_input_links()
        self.refresh()
        self.commit("Refresh")
        assert not self.director.use_git or self.git_is_clean()

    def build(self, **kwargs) -> None:
        """ Run the build if necessary. """
        self.mapper.item = self.item
        logging.info("%s: check if build is necessary", self.uid)
        necessary = self.is_build_necessary()
        if necessary:
            logging.info("%s: build is necessary", self.uid)
        force: bool = kwargs.get("force", False)
        if force:
            logging.info("%s: build is forced", self.uid)
        if force or necessary:
            self.do_run()
        else:
            logging.info("%s: build is unnecessary", self.uid)

    def label(self, item: Optional[Item] = None) -> str:
        """ Get the label of the build item description. """
        if item is None:
            item = self.item
        return f"PackageItem{item.ident}"

    def reference(self) -> str:
        """ Get a reference to the build item description. """
        return self.description.get_reference(self.label(), self.item.spec)

    def clear_description(self) -> None:
        """ Clear the build item description. """
        self.description = SphinxContent()

    def _store_description(self) -> None:
        description = "\n".join(self.description)
        key = "build-description"
        if description:
            self.item[key] = description
        else:
            self.item.data.pop(key, None)

    def add_build_section(self, **kwargs) -> None:
        """ Add the build section to the content. """
        content: SphinxContent = kwargs["content"]
        with content.section(self.item.spec, label=self.label()):
            inputs = [item.reference() for item in sorted(set(self.inputs()))]
            if len(inputs) == 1:
                content.add(f"Use input {inputs[0]}.")
            else:
                content.add_list(inputs, prologue="Use the following inputs:")
            description: str | SphinxContent = self.description
            if not description:
                description = "Support the package build."
            content.add(description)

    def has_changed(self, link: Link) -> bool:
        """
        Return true, if the build item state changed with respect to the state
        of the link, otherwise false.
        """
        return link["hash"] is None or self.digest() != link["hash"]

    def needs_build(self) -> bool:
        """
        Return true, if the build item needs a build, otherwise returns false.
        """
        return False

    def is_build_necessary(self) -> bool:
        """ Return true, if the build is necessary, otherwise false. """
        necessary = False
        for link in itertools.chain(
                self.item.links_to_parents(("make", "input"),
                                           is_link_enabled=link_is_enabled),
                self.item.links_to_children("input-to",
                                            is_link_enabled=link_is_enabled)):
            if not link.item.enabled:
                logging.info("%s: input is disabled: %s", self.uid,
                             link.item.uid)
                continue
            build_item = self.director[link.item.uid]
            if link["hash"] is None:
                logging.info("%s: input is new: %s", self.uid, build_item.uid)
            if build_item.has_changed(link):
                logging.info("%s: input has changed: %s", self.uid,
                             build_item.uid)
                necessary = True
            else:
                logging.info("%s: input has not changed: %s", self.uid,
                             build_item.uid)
        for link in self.item.links_to_parents(
                "output", is_link_enabled=link_is_enabled):
            if not link.item.enabled:
                logging.info("%s: output is disabled: %s", self.uid,
                             link.item.uid)
                continue
            build_item = self.director[link.item.uid]
            if build_item.needs_build():
                logging.info("%s: output needs build: %s", self.uid,
                             build_item.uid)
                necessary = True
            else:
                logging.info("%s: output build is not mandatory: %s", self.uid,
                             build_item.uid)
        return necessary

    def is_present(self) -> bool:
        """
        Return true, if the build item state is present, otherwise false.
        """
        return True

    def discard(self) -> None:
        """ Discard the data associated with the build item.  """

    def discard_outputs(self) -> None:
        """ Discard all outputs of the build item.  """
        logging.info("%s: discard outputs", self.uid)
        for item in self.item.parents("output",
                                      is_link_enabled=link_is_enabled):
            if not item.enabled:
                logging.info("%s: output is disabled: %s", self.uid, item.uid)
                continue
            item["build-description"] = None
            build_item = self.director[item.uid]
            build_item.discard()
            build_item.clear_description()

    def clear(self) -> None:
        """ Clear the state of the build item.  """

    def lazy_verify(self) -> bool:
        """ Lazily verify the state of the build item.  """
        return True

    def refresh(self) -> None:
        """ Refresh the build item state.  """

    def commit(self, reason: str) -> None:
        """ Commit the build item state.  """
        logging.info("%s: commit: %s", self.uid, reason)
        files = [self.item.file]
        for item in self.item.children("input-to"):
            item.save()
            files.append(item.file)
        self._store_description()
        self.item.save()
        self.git_add(files)
        self.git_commit(reason)

    def digest(self) -> str:
        """ Return the digest of the build item. """
        data = copy.deepcopy(self.item.data)
        for link in data["links"]:
            if link["role"] == "input-to":
                link["hash"] = None
        for link in self.item.links_to_children("input-to"):
            data["links"].append(link.data)
        data.pop("workspace-hash", None)
        return data_digest(data)

    def refresh_link(self, link: Link) -> None:
        """ Refresh the link to reflect the state of the build item. """
        link["hash"] = self.digest()

    def refresh_outputs(self) -> None:
        """ Refresh all outputs of the build item.  """
        logging.info("%s: refresh outputs", self.uid)
        for item in self.item.parents("output"):
            uid = item.uid
            build_item = self.director[uid]
            build_item.refresh_make_and_input_links()
            build_item.refresh()
            build_item.commit("Refresh output")

    def refresh_make_and_input_links(self) -> None:
        """ Refresh all make and input links of the build item.  """
        logging.info("%s: refresh make and input links", self.uid)
        for link in itertools.chain(self.item.links_to_parents("make"),
                                    _get_input_links(self.item)):
            self.director[link.item.uid].refresh_link(link)

    def _make(self, parent_links: Iterator[Link], child_links: Iterator[Link],
              make_link: Link) -> None:
        """ Make the outputs using the linked item.  """
        # Create a temporary item using duplicated data to make sure the maker
        # item uses no persistent state
        item_cache = self.item.cache
        item = Item(item_cache, make_link.item.uid,
                    copy.deepcopy(make_link.item.data))
        item["make-params"] = make_link["params"]
        item.init_parents(item_cache)
        # The maker item has no links to children by itself
        for link in parent_links:
            item.add_link_to_parent(link)
        for link in child_links:
            item.add_link_to_child(link)
        item.add_link_to_parent(
            Link(
                self.item, {
                    "name": make_link["output-name"],
                    "role": "output",
                    "uid": self.item.uid
                }))
        item_cache.type_provider.set_type(item)
        director = self.director
        maker = director.factory.create(director, item)
        maker.run()
        self.description.add(maker.description)

    def run(self) -> None:
        """ Run the build item tasks. """
        item = self.item
        for make_link in item.links_to_parents("make"):
            self._make(item.links_to_parents(("input", "output")),
                       item.links_to_children("input-to"), make_link)

    def substitute(self, data: Any, item: Optional[Item] = None) -> Any:
        """
        Recursively substitute the data using the item mapper of the build
        step.
        """
        if item is None:
            item = self.item
        return self.mapper.substitute_data(data, item)

    def input(self, name: str) -> "BuildItem":
        """ Return the first input associated with the name. """
        for link in _get_input_links(self.item):
            if link["name"] == name:
                return self.director[link.item.uid]
        raise KeyError

    def inputs(self, name: Optional[str] = None) -> Iterator["BuildItem"]:
        """ Yield the inputs associated with the name. """
        for link in _get_input_links(self.item):
            if name is None or link["name"] == name:
                yield self.director[link.item.uid]

    def input_link(self,
                   name: Optional[str] = None) -> tuple[Link, "BuildItem"]:
        """ Return the first link and input pair associated with the name. """
        for link in _get_input_links(self.item):
            if link["name"] == name:
                return link, self.director[link.item.uid]
        raise KeyError

    def input_links(
            self,
            name: Optional[str] = None) -> Iterator[tuple[Link, "BuildItem"]]:
        """ Yield the link and input pairs associated with the name. """
        for link in _get_input_links(self.item):
            if name is None or link["name"] == name:
                yield link, self.director[link.item.uid]

    def input_link_by_key(self, name: str, key: str,
                          value: str) -> tuple[Link, "BuildItem"]:
        """
        Get the input associated with the name and a key with the value.
        """
        for link in _get_input_links(self.item):
            if name is None or link["name"] == name and link[key] == value:
                return link, self.director[link.item.uid]
        raise ValueError(f"{self.uid}: has no {name} input with attribute "
                         f"{key} equal to {value}")

    def output(self, name: str) -> "BuildItem":
        """
        Return the first directory state production associated with the
        name.
        """
        for link in self.item.links_to_parents(
                "output", is_link_enabled=link_is_enabled):
            if link["name"] == name:
                if link.item.enabled:
                    return self.director[link.item.uid]
                logging.info("%s: output is disabled: %s", self.uid,
                             link.item.uid)
                raise KeyError
        raise KeyError

    def git_add(self, what: Union[str, Iterable[str]]) -> None:
        """ Add the files to Git. """
        if not self.director.use_git:
            return
        directory = self.git_directory
        what = to_iterable(what)
        submodules = self.director.submodules
        if submodules:
            what = [path for path in what if not path.startswith(submodules)]
        what = [os.path.relpath(path, directory) for path in what]
        if what:
            command = [
                "git", "add", "--force", "--pathspec-from-file=-",
                "--pathspec-file-nul"
            ]
            logging.info(
                "%s: add files %s to Git by running in '%s' "
                "with file list provided by stdin: %s", self.uid, what,
                directory, " ".join(command))
            subprocess.run(command,
                           check=True,
                           input="\0".join(what).encode("utf-8"),
                           cwd=directory)

    def git_is_clean(self) -> bool:
        """
        Return true, if the Git repository is clean, otherwise false.
        """
        directory = self.git_directory
        stdout: list[str] = []
        status = run_command(["git", "status", "--short"],
                             directory,
                             stdout=stdout)
        assert status == 0
        return bool(not stdout)

    def git_commit(self, reason: str) -> None:
        """ Commit the staged files to Git. """
        if not self.director.use_git:
            return
        directory = self.git_directory
        if _git_has_staged_files(directory):
            status = run_command(
                ["git", "commit", "-m", f"{self.item.uid}: {reason}"],
                directory)
            assert status == 0

    def _get_component(self, ctx: ItemGetValueContext) -> Any:
        if isinstance(self, PackageComponent):
            component = self
        else:
            assert self._component_stack[-1] is not None
            component = self._component_stack[-1]
        while True:
            try:
                args = "" if ctx.args is None else f":{ctx.args}"
                item, key_path, value = ctx.mapper.map(
                    f".:/{ctx.remaining_path}{args}", component.item)
            except ValueError:
                pass
            else:
                ctx.item = item
                ctx.path = key_path
                ctx.key_index = ""
                ctx.remaining_path = ""
                return value
            try:
                component = component.component
            except StopIteration as err:
                raise KeyError from err

    def _get_input(self, ctx: ItemGetValueContext) -> Any:
        return _get_input_or_output(ctx, self.input)

    def _get_output(self, ctx: ItemGetValueContext) -> Any:
        return _get_input_or_output(ctx, self.output)


class PackageComponent(BuildItem):
    """ Represents a package component. """

    def __init__(self,
                 director: "PackageBuildDirector",
                 item: Item,
                 mapper: None | BuildItemMapper = None) -> None:
        super().__init__(director, item, mapper)
        self.selection = ItemSelection(self.item.cache,
                                       self.item["enabled-set"],
                                       self.is_item_enabled)
        try:
            component = self.component
        except StopIteration:
            view = item.cache.top_view
        else:
            view = ItemView(component.view)
        self.view: ItemView = view

    def __getitem__(self, key: str) -> Any:
        component = self
        while True:
            if key in component.item:
                return component.substitute(component.item[key])
            try:
                component = component.component
            except StopIteration as err:
                raise KeyError(
                    f"{self.uid}: "
                    f"cannot get component value for '{key}': {err}") from err

    def run(self) -> None:
        self.description.add("Provide settings for the component.")

    def is_item_enabled(self, enabled_set: EnabledSet, item: Item) -> bool:
        """
        Return true, if the item is enabled for the enabled set, otherwise
        false.

        Where the item belongs to a component, the enabled set of the component
        is used.
        """
        try:
            component_item = build_item_input(item, "component")
        except KeyError:
            return is_enabled(enabled_set, item["enabled-by"])
        component = self.director[component_item.uid]
        assert isinstance(component, PackageComponent)
        enabled = is_enabled(component.selection.enabled_set,
                             item["enabled-by"])
        if not enabled:
            logging.info("%s: is disabled", item.uid)
        return enabled


class GenericPackageComponent(PackageComponent):
    """ Represents a generic package component. """

    def __init__(self,
                 director: "PackageBuildDirector",
                 item: Item,
                 mapper: None | BuildItemMapper = None) -> None:
        super().__init__(director, item, mapper)

        # Augment view with glossary terms
        with self.item.cache.selection(self.selection):
            with self.item.cache.view_scope(self.view):
                for glossary in self.item.parents("glossary"):
                    augment_glossary_terms(glossary, [])


class Redirection(BuildItem):
    """ Provides a build item redirection. """

    def target(self) -> BuildItem:
        """ Return the target build item of the redirection. """
        return self.input("redirection-target")


_ExportData = Callable[[Item, bool, bool], Any]


def export_data(item: Item, _present: bool, built_later: bool) -> Any:
    """ Export the item data using the presence and built later indicator. """
    data = copy.deepcopy(item.data)
    data.pop("workspace-hash", None)
    if built_later:
        for link in data["links"]:
            if link["role"] in ("input", "input-to"):
                link["hash"] = None
    return data


class BuildItemFactory:
    """
    Provides a factory to create build items for registered build item types.
    """

    def __init__(self) -> None:
        """ Initialize the dictionary of build steps """
        self.constructors: dict[str, Type[BuildItem]] = {}
        self._export_data: dict[str, _ExportData] = {}
        self._get_values: list[tuple[str, ItemGetValue]] = []

    def add_constructor(self, type_name: str, cls: Type[BuildItem]) -> None:
        """ Associate the build item constructor with the type name. """
        self.constructors[type_name] = cls
        cls.prepare_factory(self, type_name)

    def add_export_data(self, type_name: str,
                        export_data_method: _ExportData) -> None:
        """ Associate the export data method with the type name. """
        self._export_data[type_name] = export_data_method

    def add_get_value(self, type_path_key: str,
                      get_value: ItemGetValue) -> None:
        """ Add the get value method for the type key path. """
        self._get_values.append((type_path_key, get_value))

    def add_get_values_to_mapper(self, mapper: ItemMapper) -> None:
        """ Add the registered get value methods to the mapper. """
        for type_path_key, get_value in self._get_values:
            mapper.add_get_value(type_path_key, get_value)

    def create(self, director: "PackageBuildDirector",
               item: Item) -> BuildItem:
        """
        Create a build item for the item.

        The new build item will be assocated with the build director.
        """
        return self.constructors.get(item.type, BuildItem)(director, item)

    def export_data(self, item: Item, present: bool, build_order: int) -> Any:
        """ Export the item data using the presence indicator. """
        built_later = item.view.get("package-build-order",
                                    len(item.cache)) > build_order
        return self._export_data.get(item.type, export_data)(item, present,
                                                             built_later)


def _match(uid: str, patterns: Optional[list[str]]) -> bool:
    if patterns is None:
        return False
    return any(fnmatch.fnmatch(uid, pattern) for pattern in patterns)


def _gather_build_uids_of_package(item: Item, build_uids: set[str]) -> None:
    assert not item.is_proxy()
    if item.uid in build_uids:
        return
    build_uids.add(item.uid)
    with item.cache.selection(item.view["component"].selection):
        for dependency in itertools.chain(item.children("input"),
                                          item.parents("input-to")):
            _gather_build_uids_of_package(dependency, build_uids)


class PackageBuildDirector(dict):
    """
    The package build director contains the package build state and runs the
    build.
    """

    def __init__(self,
                 item_cache: ItemCache,
                 package_uid: str,
                 factory: BuildItemFactory,
                 use_git: Optional[bool] = False) -> None:
        self.item_cache = item_cache
        self.package_uid = package_uid
        self.factory = factory
        self.use_git = use_git
        self.submodules: tuple[str, ...] = tuple()
        item_cache.top_view.add_get_missing("component", self._get_component)

    def __missing__(self, uid: str) -> BuildItem:
        logging.info("%s: create build item", uid)
        return self.setdefault(uid,
                               self.factory.create(self, self.item_cache[uid]))

    @property
    def package(self) -> PackageComponent:
        """ Is the package component. """
        build_item = self[self.package_uid]
        assert isinstance(build_item, PackageComponent)
        return build_item

    def remove(self, uid: str) -> None:
        """
        Remove the build item associated with the UID from the director and
        the item cache.
        """
        self.pop(uid, None)
        self.item_cache.remove_item(uid)

    def add_submodule(self, submodule: str) -> None:
        """ Adds the submodule. """
        submodule = f"{submodule.rstrip('/')}/"
        self.submodules = self.submodules + (submodule, )

    def _get_component(self, item: Item) -> PackageComponent:
        try:
            component_item = build_item_input(item, "component")
        except KeyError:
            return self.package
        component = self[component_item.uid]
        assert isinstance(component, PackageComponent)
        return component

    def _gather_build_dependencies(self, item: Item,
                                   deps: dict[str, set[str]]) -> None:
        assert not item.is_proxy()
        uid = item.uid
        if uid in deps:
            return
        deps[uid] = set()
        component = item.view["component"]
        with self.item_cache.selection(component.selection):
            for dependency in itertools.chain(
                    item.parents(
                        ("input", "make", "weak-package-build-dependency")),
                    item.children(("input-to", "output"))):
                deps[uid].add(dependency.uid)
                self._gather_build_dependencies(dependency, deps)

    def _gather_ordered_build_uids_of_package(
            self, only: Optional[list[str]]) -> list[str]:
        item_cache = self.item_cache
        package = self.package.item
        build_uids: set[str] = set()
        _gather_build_uids_of_package(package, build_uids)
        if only is not None:
            build_only: set[str] = set()
            for uid in build_uids:
                if _match(uid, only):
                    build_only.add(uid)
            build_uids = build_only
        deps: dict[str, set[str]] = {}
        for uid in sorted(build_uids):
            self._gather_build_dependencies(item_cache[uid], deps)
        ordered_build_uids = list(
            graphlib.TopologicalSorter(deps).static_order())
        for uid in ordered_build_uids:
            logging.info("%s: depends on: %s", uid, sorted(deps[uid]))
        logging.info("%s: build order: %s", package.uid, ordered_build_uids)
        return ordered_build_uids

    def _build(self, method: str, **kwargs):
        only: Optional[list[str]] = kwargs.get("only", None)
        force: list[str] = kwargs.get("force", [])
        skip: Optional[list[str]] = kwargs.get("skip", None)
        item_cache = self.item_cache
        package = self.package
        with item_cache.selection(package.selection):
            build_uids = self._gather_ordered_build_uids_of_package(only)
            for build_order, uid in enumerate(build_uids):
                item = item_cache[uid]
                item.view["package-build-order"] = build_order
                component = item.view["component"]
                logging.info("%s: use component %s", uid, component.uid)
                with item_cache.selection(component.selection):
                    with item_cache.view_scope(component.view):
                        # Construct the builder even if building it is skipped
                        builder = self[uid]
                        is_forced = _match(uid, force)
                        if _match(uid, skip) and not is_forced:
                            logging.info(
                                "%s: build is skipped due to skip filter", uid)
                            continue
                        kwargs["component"] = component
                        kwargs["force"] = is_forced
                        getattr(builder, method)(**kwargs)

    def build_package(self,
                      only: Optional[list[str]] = None,
                      force: Optional[list[str]] = None,
                      skip: Optional[list[str]] = None) -> None:
        """ Build the package. """
        logging.info("%s: build the package", self.package_uid)
        self._build("build", only=only, force=force, skip=skip)
        logging.info("%s: finished building package", self.package_uid)

    def _build_only(self, item: Item, force_patterns: Optional[list[str]],
                    forced: set[str], seen: set[str],
                    build_order: int) -> None:
        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-positional-arguments
        if item.uid in seen:
            return
        if _match(item.uid, force_patterns):
            forced.add(item.uid)
        is_forced = item.uid in forced
        logging.info("%s: prepare %sbuild", item.uid,
                     "forced " if is_forced else "")
        seen.add(item.uid)
        is_output = False
        for link in itertools.chain(
                item.links_to_children(("input-to", "output")),
                item.links_to_parents("input")):
            if link.role == "output":
                is_output = True
                if is_forced:
                    forced.add(link.item.uid)
            self._build_only(link.item, force_patterns, forced, seen,
                             build_order - 1)
        if not is_output:
            item.view["package-build-order"] = build_order
            component = item.view["component"]
            logging.info("%s: use component %s", item.uid, component.uid)
            with self.item_cache.selection(component.selection):
                with self.item_cache.view_scope(component.view):
                    build_item = self[item.uid]
                    is_necessary = build_item.is_build_necessary()
                    if is_necessary:
                        logging.info("%s: build is necessary", item.uid)
                    if is_forced:
                        logging.info("%s: build is forced", item.uid)
                    if is_necessary or is_forced:
                        for dependency in itertools.chain(
                                item.children("input-to"),
                                item.parents("input")):
                            dependency_build_item = self[dependency.uid]
                            if not dependency_build_item.is_present():
                                logging.info(
                                    "%s: dependency is not present: %s",
                                    item.uid, dependency.uid)
                                forced.add(dependency.uid)
                                self._build_only(dependency, force_patterns,
                                                 forced, set(),
                                                 build_order - 1)
                        build_item.do_run()
        logging.info("%s: building done", item.uid)

    def build_only(self,
                   uids: str | list[str],
                   force_patterns: Optional[list[str]] = None) -> None:
        """
        Build the items associated with the UIDs and their dependencies if
        necessary.
        """
        with self.item_cache.selection(self.package.selection):
            seen: set[str] = set()
            for uid in to_iterable(uids):
                self._build_only(self.item_cache[uid], force_patterns, set(),
                                 seen, 0)

    def add_build_description(self, content: SphinxContent,
                              build_uids: list[str]) -> None:
        """ Add the build description of the build items to the content. """
        self._build("add_build_section", content=content, only=build_uids)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides a test runner using the ESA run tests service.  """

# Copyright (C) 2023, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import bz2
import contextlib
import datetime
import logging
import math
from pathlib import Path
import socket
import time
from typing import Iterator
import yaml

from specitems import Item
from specware import run_command

from .directorystate import RepositoryState
from .pkgitems import PackageBuildDirector
from .testrunner import Executable, Report, TestRunner


class ESATestRunner(TestRunner):
    """ Runs tests using the ESA run tests service. """

    def __init__(self, director: PackageBuildDirector, item: Item) -> None:
        super().__init__(director, item)
        self._request_policies = {
            "limit-executable-count": self._get_request_limit_executable_count,
            "use-timeouts": self._get_request_use_timeouts,
        }

    def _target_board(self) -> str:
        return self["bsp-to-target-board"][self.substitute(
            "${.:/component/arch}/${.:/component/bsp}")]

    def describe(self) -> str:
        return f"""This test procedure requests the `ESA run tests service
<https://gitlab.esa.int/flight-software/fsw-boards-server/\
FSWBoardsServerRemoteExecution_Binaries.git>`__
to run the test programs on the ``{self._target_board()}`` target board."""

    def _create_branch(self, working_directory: Path,
                       repository: RepositoryState) -> str:
        """
        Create a unique branch to carry out the request and return its name.
        """
        hostname = socket.gethostname()
        now = datetime.datetime.utcnow().strftime('%Y-%m-%d-%H-%M-%S-%f')
        branch = f"pkg-{self.component['name']}-{hostname}-{now}"
        status = run_command(
            ["git", "checkout", "-B", branch, repository["commit"]],
            cwd=str(working_directory))
        assert status == 0
        return branch

    @contextlib.contextmanager
    def _temporary_branch(self, working_directory: Path,
                          repository: RepositoryState) -> Iterator[str]:
        """
        Context manager to create and clean up a temporary Git branch.
        """
        branch = self._create_branch(working_directory, repository)
        yield branch
        if self["delete-remote-branch"]:
            status = run_command(["git", "push", "-d", "origin", branch],
                                 cwd=str(working_directory))
            assert status == 0
        # Return to original branch
        status = run_command(["git", "checkout", repository["branch"]],
                             cwd=str(working_directory))
        assert status == 0
        # Delete temporary branch locally
        status = run_command(["git", "branch", "-D", branch],
                             cwd=str(working_directory))
        assert status == 0

    def _copy_and_prepare_executables(
            self, working_directory: Path,
            executables: list[Executable]) -> list[Executable]:
        """
        Copy, strip, and compress (bzip2) the executables into the working
        directory.
        """
        remaining: list[Executable] = []
        strip = self["strip-program-path"]
        for exe in executables:
            source = Path(exe.path)
            stripped = working_directory / source.name
            status = run_command(
                [strip, str(source), "-o",
                 str(stripped)],
                cwd=str(working_directory))
            assert status == 0
            data = bz2.compress(stripped.read_bytes())
            stripped.unlink()
            name = f"{source.name}.bz2"
            remaining.append(Executable(name, exe.digest, exe.timeout))
            compressed = working_directory / name
            compressed.write_bytes(data)
        return remaining

    def _get_request_limit_executable_count(
            self, remaining: list[Executable],
            policy: dict) -> tuple[dict, int, list[Executable]]:
        max_executable_count = policy["max-executable-count"]
        todo = remaining[:max_executable_count]
        remaining[:] = remaining[max_executable_count:]
        overall_timeout = sum(int(math.ceil(exe.timeout)) for exe in todo)
        jobs = [{exe.path: None} for exe in todo]
        return {
            "version":
            1,
            "target_board":
            self._target_board(),
            "timeout_in_seconds":
            min(overall_timeout, self["max-overall-timeout-in-seconds"]),
            "jobs":
            jobs,
        }, overall_timeout, todo

    def _get_request_use_timeouts(
            self, remaining: list[Executable],
            _policy: dict) -> tuple[dict, int, list[Executable]]:
        overall_timeout = 0
        todo: list[Executable] = []
        max_overall_timeout = self["max-overall-timeout-in-seconds"]
        jobs: list[dict[str, dict[str, int]]] = []
        max_timeout = 0
        while remaining:
            exe = remaining.pop(0)
            timeout = int(math.ceil(exe.timeout))
            if overall_timeout == 0:
                # Ignore maximum overall timeout for first executable
                overall_timeout = timeout
            else:
                new_overall_timeout = overall_timeout + timeout
                if new_overall_timeout > max_overall_timeout:
                    remaining.insert(0, exe)
                    break
                overall_timeout = new_overall_timeout
            max_timeout = max(max_timeout, timeout)
            todo.append(exe)
            jobs.append({exe.path: {"timeout_in_seconds": timeout}})
        return {
            "version": 1,
            "target_board": self._target_board(),
            "timeout_in_seconds": max_timeout,
            "jobs": jobs,
        }, overall_timeout, todo

    def _write_request_file(
        self,
        working_directory: Path,
        remaining: list[Executable],
    ) -> tuple[str, int, list[Executable]]:
        """
        Write the request file to run the prepared remaining executables.
        Returns (request filename, overall timeout, to do executables).
        """
        policy = self["request-policy"]
        request, overall_timeout, todo = self._request_policies[
            policy["policy"]](remaining, policy)
        request_file = working_directory / "request.yaml"
        request_file.write_text(
            yaml.dump(request, default_flow_style=False, allow_unicode=True),
            encoding="utf-8",
        )
        return request_file.name, overall_timeout, todo

    def _wait_for_update(self, working_directory: Path, branch: str,
                         overall_timeout: int) -> None:
        """
        Wait for a branch update on the remote repository.
        """
        polling_interval = self["git-fetch-polling-interval-in-seconds"]
        max_duration = float(self["max-request-delay-in-seconds"] +
                             overall_timeout)
        begin = time.monotonic()
        while time.monotonic() - begin < max_duration:
            time.sleep(polling_interval)
            status = run_command(["git", "fetch", "origin", branch],
                                 cwd=str(working_directory))
            if status != 0:
                logging.warning("%s: fetch from remote repository failed",
                                self.uid)
            stdout: list[str] = []
            status = run_command(
                ["git", "log", f"HEAD..origin/{branch}", "--oneline"],
                stdout=stdout,
                cwd=str(working_directory),
            )
            assert status == 0
            if stdout:
                return

    def _add_reports(self, working_directory: Path,
                     executables: list[Executable],
                     reports: list[Report]) -> None:
        for exe in executables:
            result_path = working_directory / f"{exe.path}.result.txt"
            report: Report = {
                "executable": exe.path,
                "executable-sha512": exe.digest,
                "command-line": [],
            }
            try:
                output = result_path.read_text(encoding="utf-8")
            except FileNotFoundError:
                output = ""
                report["error"] = "waiting for response timed out"
            report["output"] = output.rstrip().replace("\r\n",
                                                       "\n").split("\n")
            reports.append(report)

    def run_tests(self, executables: list[Executable]) -> list[Report]:
        reports: list[Report] = []
        repository = self.director[self.item.parent(
            "weak-package-build-dependency").uid]
        assert isinstance(repository, RepositoryState)
        working_directory = Path(repository.directory)
        counter = 0
        with self._temporary_branch(working_directory, repository) as branch:
            remaining = self._copy_and_prepare_executables(
                working_directory, executables)
            while remaining:
                request_file, overall_timeout, todo = self._write_request_file(
                    working_directory, remaining)
                files = [exe.path for exe in todo]
                status = run_command(["git", "add"] + files + [request_file],
                                     cwd=str(working_directory))
                assert status == 0
                status = run_command(
                    ["git", "commit", "-m", f"Request {counter}"],
                    cwd=str(working_directory),
                )
                assert status == 0
                status = run_command(["git", "push", "origin", branch],
                                     cwd=str(working_directory))
                assert status == 0
                self._wait_for_update(working_directory, branch,
                                      overall_timeout)
                status = run_command(
                    ["git", "reset", "--hard", f"origin/{branch}"],
                    cwd=str(working_directory),
                )
                assert status == 0
                self._add_reports(working_directory, todo, reports)
                counter += 1
        return reports



# SPDX-License-Identifier: BSD-2-Clause
""" Builds documents presenting specification items. """

# Copyright (C) 2022, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=too-many-lines

import functools
import itertools
import os
import re
from typing import Any, Callable, Iterable, NamedTuple

from specitems import (COL_SPAN, EnabledSet, GenericContent, Item,
                       ItemGetValueContext, ItemMapper, ROW_SPAN, is_enabled,
                       Link, link_is_enabled, make_label, SphinxContent)
from specware import (CodeMapper, TransitionMap, PreCondsOfPostCond,
                      align_declarations, document_directive, document_option,
                      forward_declaration)

from .docbuilder import DocumentBuilder
from .pkgitems import PackageBuildDirector
from .linkhub import get_kind, spec_label, SpecMapper
from .perfimages import environment_order
from .rtems import RTEMSItemCache
from .speccompare import CompareSpecsRegistry
from .testaggregator import get_test_result_status
from .util import duration


class _Context(NamedTuple):
    content: SphinxContent
    item: Item
    mapper: SpecMapper
    code_mapper: ItemMapper
    spec: RTEMSItemCache
    file_path: str


def _add_text(ctx: _Context, key: str, name: str) -> None:
    text = ctx.item.get(key, None)
    if text:
        ctx.content.add_rubric(f"{name}:")
        ctx.content.wrap(ctx.mapper.substitute(text))


def _add_sdd_link(ctx: _Context) -> None:
    link = ctx.item.view["document-links"].get("sdd", None)
    if link:
        ctx.content.add_rubric("SOFTWARE DESIGN:")
        kind = get_kind(ctx.item)
        ctx.content.add(
            f"This {kind} is realised by the software design element {link}.")


def _add_links(
        ctx: _Context,
        role: str | list[str],
        name: str,
        parent_role: str,
        child_role: str,
        child_prefix: str = "This",
        is_link_enabled: Callable[[Link], bool] = link_is_enabled) -> None:
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    # pylint: disable=too-many-locals
    parents = [
        link.item for link in ctx.item.links_to_parents(role)
        if is_link_enabled(link)
    ]
    children = [
        link.item for link in ctx.item.links_to_children(role)
        if is_link_enabled(link)
    ]
    if parents or children:
        get_link = ctx.mapper.get_link
        plural = "S" if len(parents) + len(children) > 1 else ""
        ctx.content.add_rubric(f"{name}{plural}:")
        kind = get_kind(ctx.item)
        if len(parents) == 1:
            parent = parents[0]
            parent_kind = get_kind(parent)
            ctx.content.wrap(f"This {kind} {parent_role} the "
                             f"{parent_kind} {get_link(parent)}.")
        elif len(parents) > 1:
            ctx.content.add_list(
                [get_link(parent) for parent in parents],
                f"This {kind} {parent_role} the following items:")
        if len(children) == 1:
            child = children[0]
            child_kind = get_kind(child)
            ctx.content.wrap(f"{child_prefix} {kind} {child_role} "
                             f"{child_kind} {get_link(child)}.")
        elif len(children) > 1:
            ctx.content.add_list(
                [get_link(child) for child in children],
                f"{child_prefix} {kind} {child_role} the following items:")


def _add_default_links(ctx: _Context) -> None:
    _add_sdd_link(ctx)
    _add_links(ctx, "interface-enumerator", "ENUMERATOR", "provides",
               "is provided by")
    _add_links(ctx, "requirement-refinement", "REFINEMENT", "refines",
               "is refined by")
    _add_links(ctx, ["interface-ingroup", "interface-ingroup-hidden"],
               "GROUP MEMBERSHIP", "is a member of", "contains")
    _add_links(ctx, "interface-placement", "INTERFACE PLACEMENT",
               "is placed into", "contains")
    _add_links(ctx,
               "interface-function",
               "INTERFACE FUNCTION",
               "specifies the function of",
               "is specified by the",
               child_prefix="The function of this")
    _add_links(ctx, "function-implementation", "FUNCTION IMPLEMENTATION",
               "uses functions implemented by",
               "implements a function used by")
    _add_links(ctx,
               "interface-include",
               "INTERFACE INCLUDE",
               "includes",
               "is included by",
               is_link_enabled=functools.partial(_is_include_enabled,
                                                 ctx.spec.enabled_set))


def _add_validated_items(ctx: _Context) -> None:
    _add_links(ctx, "validation", "VALIDATED ITEM", "validates", "oops")


def _add_code_block(content: SphinxContent, code: GenericContent) -> None:
    content.add_rubric("INTERFACE:")
    with content.directive("code-block", "c"):
        content.add(code)


def _add_type_definition(content: SphinxContent, name: str,
                         definition_kind: str, type_name: str,
                         code: list[str]) -> None:
    if definition_kind == f"{type_name}-only":
        code.insert(0, f"{type_name} {name} {{")
        code.append("};")
    elif definition_kind == "typedef-only":
        code.insert(0, f"typedef {type_name} {{")
        code.append(f"}} {name};")
    else:
        code.insert(0, f"typedef {type_name} {name} {{")
        code.append(f"}} {name};")
    _add_code_block(content, code)


def _definition(definition: dict, mapper: ItemMapper,
                enabled_set: EnabledSet) -> Any:
    enabled_set = frozenset(f"defined({enable})" for enable in enabled_set)
    for variant in definition["variants"]:
        if is_enabled(enabled_set,
                      mapper.substitute_data(variant["enabled-by"])):
            return variant["definition"]
    return definition["default"]


def _item_definition(item: Item, mapper: ItemMapper,
                     enabled_set: EnabledSet) -> str:
    return mapper.substitute(
        _definition(item["definition"], mapper, enabled_set)).strip()


def _document_unspecified(ctx: _Context,
                          prefix: str = "",
                          postfix: str = "") -> None:
    ctx.content.add_rubric("REQUIREMENT:")
    kind = get_kind(ctx.item)
    container = ctx.item.parent('interface-placement')
    container_kind = get_kind(container)
    ctx.content.wrap(
        f"The {ctx.mapper.get_link(container)} {container_kind} "
        f"shall provide the {kind} ``{prefix}{ctx.item['name']}{postfix}``.")


def _document_unspecified_type(ctx: _Context) -> None:
    type_kind = ctx.item['interface-type'][12:]
    _document_unspecified(ctx, prefix=f"{type_kind} ")
    _add_code_block(ctx.content, f"{type_kind} {ctx.item['name']} {{ ... }};")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_unspecified_define(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_code_block(ctx.content, f"#define {ctx.item['name']} ...")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_unspecified_enumerator(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_code_block(
        ctx.content,
        ["enum ... {", "  ....", f"  {ctx.item['name']} ...", "  ....", "};"])
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_unspecified_function(ctx: _Context) -> None:
    _document_unspecified(ctx, postfix="()")
    _add_code_block(ctx.content, f"... {ctx.item['name']}( ... );")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_unspecified_group(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_unspecified_object(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_code_block(ctx.content, f"extern ... {ctx.item['name']} ...;")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_unspecified_typedef(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_code_block(ctx.content, f"typedef ... {ctx.item['name']} ...;")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_constraint(ctx: _Context) -> None:
    _add_text(ctx, "text", "CONSTRAINT")
    _add_text(ctx, "rationale", "RATIONALE")
    _add_default_links(ctx)
    _add_links(ctx, "constraint", "CONSTRAINT ITEM", "constrained by",
               "is applicable to")


def _document_requirement(ctx: _Context) -> None:
    _add_text(ctx, "text", "REQUIREMENT")
    _add_text(ctx, "rationale", "RATIONALE")
    _add_default_links(ctx)
    _add_validations(ctx)


def _add_perf_measurement(ctx: _Context,
                          test_results: list[dict[str, Any]]) -> None:
    for measurement_data in test_results:
        test_case = measurement_data["test-case"]
        test_suite = test_case["test-suite"]
        config_data = test_suite["config"]
        test_case_item = ctx.item.cache[test_case["uid"]]
        test_suite_item = ctx.item.cache[test_suite["uid"]]
        ctx.content.wrap(f"""For the configuration
``{config_data['name']}``, the test case `{test_case_item.spec_2}
<{test_suite['link']}>`__ of the test suite `{test_suite_item.spec_2}
<{test_case['link']}>`__ reported the following `runtime performance
measurements <{measurement_data['link']}>`__:""")
        ctx.content.add_image(
            os.path.relpath(f"{measurement_data['boxplot']}.*",
                            os.path.dirname(ctx.file_path)), "50%")


def _document_perf_runtime(ctx: _Context) -> None:
    _add_text(ctx, "text", "REQUIREMENT")
    ctx.content.add_rubric("RUNTIME PERFORMANCE LIMITS:")
    for link in ctx.item.links_to_children("performance-runtime-limits"):
        for target in link.item.children(
                "performance-runtime-limits-provider"):
            ctx.content.wrap(f"""For the
{ctx.mapper.get_link(target)} target, the following runtime
performance limits shall apply:""")
            rows: list[tuple[str | int, ...]] = [
                ("${ENVIRONMENT}", "${LIMIT_KIND}", "${LIMIT_CONDITION}")
            ]
            for env, limits in sorted(link["limits"].items(),
                                      key=environment_order):
                env_name, _, _ = env.partition("/")
                env_item = ctx.spec.name_to_item[
                    f"perf-runtime-env/{env_name}"]
                lower_bound = limits["min-lower-bound"]
                rows.append(
                    (f":ref:`{env} <{make_label(env_item.spec)}>`", "Minimum",
                     f"{duration(lower_bound)} :math:`\\leq` Minimum"))
                lower_bound = limits["median-lower-bound"]
                upper_bound = limits["median-upper-bound"]
                rows.append((COL_SPAN, "Median",
                             f"{duration(lower_bound)} :math:`\\leq` Median "
                             f":math:`\\leq` {duration(upper_bound)}"))
                upper_bound = limits["max-upper-bound"]
                rows.append((COL_SPAN, "Maximum",
                             f"Maximum :math:`\\leq` {duration(upper_bound)}"))
            ctx.content.add_grid_table(rows, [25, 25, 50])
            try:
                test_results = ctx.item.view["test-results"][target.uid]
            except KeyError:
                with ctx.content.directive("warning"):
                    ctx.content.wrap("""There are no runtime measurements
available for this requirement.""")
            else:
                _add_perf_measurement(ctx, test_results)
    _add_text(ctx, "rationale", "RATIONALE")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_perf_runtime_env(ctx: _Context) -> None:
    _document_requirement(ctx)
    ctx.content.add_rubric("NAME:")
    ctx.content.wrap(f"""The RTEMS Test Framework name of this runtime
measurement environment is ``{ctx.item['name']}``.""")


_UPPER = re.compile(r"[A-Z]+")


def _upper_spacify(match: re.Match[str]) -> str:
    return f"\u200b{match.group(0)}"


def _spacify(name: str) -> str:
    return _UPPER.sub(_upper_spacify, name).lstrip("\u200b")


def _name_ref(ctx: _Context, name: str, which: str) -> str:
    label = make_label(f"{ctx.item.spec} {which} {name}")
    return f":ref:`{_spacify(name)} <{label}>`"


def _add_pre_condition_variants(ctx: _Context, transition_map: TransitionMap,
                                pre_conds: Any) -> None:
    for row in pre_conds:
        entries = []
        for co_idx, co_states in enumerate(row):
            co_name = transition_map.pre_co_idx_to_co_name(co_idx)
            co_name = _name_ref(ctx, co_name, "Pre")
            states = [
                transition_map.pre_co_idx_st_idx_to_st_name(co_idx, st_idx)
                for st_idx in set(co_states)
            ]
            if len(states) == 1:
                entries.append(f"{co_name} = {states[0]}")
            else:
                entries.append(f"{co_name} = {{{', '.join(states)}}}")
        ctx.content.add("")
        ctx.content.add("    * " + ", ".join(entries))


def _states(transition_map: TransitionMap, co_idx: int,
            states: list[int]) -> str:
    return ", ".join(
        _spacify(transition_map.pre_co_idx_st_idx_to_st_name(co_idx, st_idx))
        for st_idx in set(states))


def _add_transition_map(
    ctx: _Context
) -> tuple[TransitionMap, list[tuple[str, PreCondsOfPostCond]]]:
    infeasible_pre_conds: list[tuple[str, PreCondsOfPostCond]] = []
    transition_map = TransitionMap(ctx.item, "N/A")
    rows: list[Iterable[str | int]] = [
        ("Pre-Conditions", ) + (ROW_SPAN, ) *
        (transition_map.pre_co_count - 1) + ("Post-Conditions", ) +
        (ROW_SPAN, ) * (transition_map.post_co_count - 1)
    ]
    rows.append(
        tuple(
            itertools.chain((_name_ref(ctx, condition["name"], "Pre")
                             for condition in ctx.item["pre-conditions"]),
                            (_name_ref(ctx, condition["name"], "Post")
                             for condition in ctx.item["post-conditions"]))))
    for post_co, pre_co_collection in transition_map.get_post_conditions(
            ctx.spec.enabled_set):
        if post_co[0]:
            infeasible_pre_conds.append(
                (transition_map.skip_idx_to_name(post_co[0]),
                 pre_co_collection))
            post_co_row = ((_name_ref(
                ctx, transition_map.skip_idx_to_name(post_co[0]), "Skip"), ) +
                           (ROW_SPAN, ) * (transition_map.post_co_count - 1))
            post_co_col_span: tuple[str | int,
                                    ...] = ((COL_SPAN, ) +
                                            (COL_SPAN | ROW_SPAN, ) *
                                            (transition_map.post_co_count - 1))
        else:
            post_co_row = tuple(
                _spacify(
                    transition_map.post_co_idx_st_idx_to_st_name(
                        co_idx, st_idx))
                for co_idx, st_idx in enumerate(post_co[1:]))
            post_co_col_span = (COL_SPAN, ) * transition_map.post_co_count
        for pre_co in pre_co_collection:
            pre_co_row = tuple(
                _states(transition_map, co_idx, co_states)
                for co_idx, co_states in enumerate(pre_co))
            rows.append(pre_co_row + post_co_row)
            post_co_row = post_co_col_span
    co_count = transition_map.pre_co_count + transition_map.post_co_count
    with ctx.content.latex_environment("landscape", use=co_count > 10):
        font_size = -((co_count + 4) // 5)
        with ctx.content.latex_font_size(font_size):
            cell_width = 100 // co_count
            widths = [100 - cell_width *
                      (co_count - 1)] + [cell_width] * (co_count - 1)
            ctx.content.add_grid_table(rows, widths=widths, header_rows=2)
    return transition_map, infeasible_pre_conds


def _add_conditions(ctx: _Context, which: str) -> None:
    ctx.content.add_rubric(f"{which.upper()}-CONDITIONS:")
    caption = which[0].upper() + which[1:]
    for condition in ctx.item[f"{which}-conditions"]:
        name = condition['name']
        ctx.content.add_label(make_label(f"{ctx.item.spec} {caption} {name}"))
        with ctx.content.directive("topic", f"{caption}-Condition - {name}"):
            ctx.content.add(f"""The *{name}* {which}-condition has the
following states:""")
            for state in condition["states"]:
                text = ctx.mapper.substitute(state["text"])
                ctx.content.add_definition_item(state["name"], text)


def _document_action_requirement(ctx: _Context) -> None:
    ctx.content.add_rubric("REQUIREMENT:")
    ctx.content.wrap("""The function shall be specified by the following state
transition map which defines for each feasible pre-condition state variant the
resulting post-condition state variant produced by the trigger action.""")
    _add_text(ctx, "rationale", "RATIONALE")
    _add_default_links(ctx)
    _add_validations(ctx)
    _add_conditions(ctx, "pre")
    ctx.content.add_rubric("TRIGGER ACTION:")
    ctx.content.wrap(ctx.mapper.substitute(ctx.item["text"]))
    _add_conditions(ctx, "post")
    ctx.content.add_rubric("TRANSITION MAP:")
    ctx.content.wrap("""For each of the resulting post-condition state variants
below, the set of producing pre-condition variants is listed.""")
    transition_map, infeasible_pre_conds = _add_transition_map(ctx)
    if infeasible_pre_conds:
        ctx.content.add_rubric("INFEASIBLE PRE-CONDITION VARIANTS:")
        for reason, pre_conds in infeasible_pre_conds:
            ctx.content.add_label(make_label(f"{ctx.item.spec} Skip {reason}"))
            ctx.content.add(f"{ctx.content.emphasize(reason)}:")
            ctx.content.paste(
                ctx.mapper.substitute(ctx.item["skip-reasons"][reason]))
            ctx.content.paste(
                """Therefore, the following pre-condition state variants
are infeasible:""")
            _add_pre_condition_variants(ctx, transition_map, pre_conds)


def _document_acfg_group(ctx: _Context) -> None:
    _add_text(ctx, "text", "REQUIREMENT")
    _add_text(ctx, "rationale", "RATIONALE")
    _add_text(ctx, "description", "DESCRIPTION")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_acfg_option(ctx: _Context) -> None:
    ctx.content.add_rubric("REQUIREMENT:")
    container = ctx.item.parent('interface-placement')
    container_kind = get_kind(container)
    ctx.content.wrap(
        f"The {ctx.mapper.get_link(container)} {container_kind} shall "
        f"provide the application configuration option ``{ctx.item['name']}``."
    )
    document_option(ctx.content, ctx.mapper, ctx.item, ctx.spec.enabled_set)
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_directive(ctx: _Context) -> None:
    _document_unspecified(ctx, postfix="()")
    ctx.content.add_rubric("BRIEF DESCRIPTION:")
    document_directive(ctx.content, ctx.mapper, ctx.item, ctx.spec.enabled_set)
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_register_block(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_text(ctx, "brief", "BRIEF DESCRIPTION")
    ctx.content.add_rubric("REGISTER BLOCK:")
    rows: list[tuple[str | int, ...]] = [("Offset", "Register")]
    for member in ctx.item["definition"]:
        definition = _definition(member, ctx.mapper, ctx.spec.enabled_set)
        count = definition["count"]
        array = f"[ {count} ]" if count > 1 else ""
        rows.append((f"{member['offset']:#x}", f"{definition['name']}{array}"))
    ctx.content.add_grid_table(rows, [20, 80])
    for reg in ctx.item["registers"]:
        ctx.content.add_rubric(f"REGISTER {reg['name']}:")
        rows = [(f"Bits [0:{reg['width'] - 1}]", f"{reg['brief'].strip()}")]
        for bits in reg["bits"]:
            definition = _definition(bits, ctx.mapper, ctx.spec.enabled_set)
            for field in definition:
                start = field["start"]
                width = field["width"]
                if width == 1:
                    pos = str(start)
                else:
                    pos = f"[{start}:{start + width - 1}]"
                rows.append((pos, field["name"]))
        ctx.content.add_grid_table(rows, [20, 80])
    _add_text(ctx, "description", "DESCRIPTION")
    _add_text(ctx, "notes", "NOTES")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_compound(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_text(ctx, "brief", "BRIEF DESCRIPTION")
    enabled_set = frozenset(f"defined({enable})"
                            for enable in ctx.spec.enabled_set)
    members = []
    definitions = []
    for index, member in enumerate(ctx.item["definition"]):
        prefix = f"definition[{index}]"
        for variant in member["variants"]:
            if is_enabled(enabled_set,
                          ctx.mapper.substitute_data(variant["enabled-by"])):
                prefix_2 = f"{prefix}/variants[{index}]/definition"
                definition = variant["definition"]
                break
        else:
            prefix_2 = f"{prefix}/default"
            definition = member["default"]
        if definition is not None:
            text = definition["brief"].strip()
            if definition["description"]:
                text += "\n" + definition["description"].strip()
            members.append((definition["name"],
                            ctx.mapper.substitute(text, prefix=prefix_2)))
            definitions.append(
                ctx.code_mapper.substitute(definition["definition"],
                                           prefix=prefix_2))
    decls = align_declarations(definitions)
    decls = [f"  {decl};" for decl in decls]
    _add_type_definition(ctx.content, ctx.item["name"],
                         ctx.item["definition-kind"],
                         ctx.item["interface-type"], decls)
    ctx.content.add_rubric("MEMBERS:")
    for member in members:
        ctx.content.add_definition_item(member[0], member[1])
    _add_text(ctx, "description", "DESCRIPTION")
    _add_text(ctx, "notes", "NOTES")
    _add_default_links(ctx)
    _add_validations(ctx)


_SPACE = re.compile(r"\s+")


def _enumerator(item: Item, mapper: ItemMapper,
                enabled_set: EnabledSet) -> str:
    definition = _item_definition(item, mapper, enabled_set)
    name = item["name"]
    if definition:
        definition = _SPACE.sub(" ", definition)
        return f"{name} = {definition}"
    return name


def _document_enumeration(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_text(ctx, "brief", "BRIEF DESCRIPTION")
    enumerators = [
        f"  {_enumerator(name, ctx.code_mapper, ctx.spec.enabled_set)},"
        for name in ctx.item.parents("interface-enumerator")
    ]
    _add_type_definition(ctx.content, ctx.item["name"],
                         ctx.item["definition-kind"], "enum", enumerators)
    _add_text(ctx, "description", "DESCRIPTION")
    _add_text(ctx, "notes", "NOTES")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_enumerator(ctx: _Context) -> None:
    ctx.content.add_rubric("REQUIREMENT:")
    name = ctx.item["name"]
    enum = ctx.item.child('interface-enumerator')
    ctx.content.wrap(f"The {ctx.mapper.get_link(enum)} enumeration "
                     f"shall provide the enumerator ``{name}``.")
    _add_text(ctx, "brief", "BRIEF DESCRIPTION")
    enumerator = [
        "  ...",
        f"  {_enumerator(ctx.item, ctx.code_mapper, ctx.spec.enabled_set)}",
        "  ...",
    ]
    _add_type_definition(ctx.content, enum["name"], enum["definition-kind"],
                         "enum", enumerator)
    _add_text(ctx, "description", "DESCRIPTION")
    _add_text(ctx, "notes", "NOTES")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_define(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_text(ctx, "brief", "BRIEF DESCRIPTION")
    _add_code_block(
        ctx.content, f"#define {ctx.item['name']} "
        f"{_item_definition(ctx.item, ctx.code_mapper, ctx.spec.enabled_set)}")
    _add_text(ctx, "description", "DESCRIPTION")
    _add_text(ctx, "notes", "NOTES")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_domain(ctx: _Context) -> None:
    ctx.content.add_rubric("REQUIREMENT:")
    ctx.content.wrap(
        f"There shall be the interface domain ``{ctx.item['name']}``.")
    _add_text(ctx, "description", "DESCRIPTION")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_group(ctx: _Context) -> None:
    ctx.content.add_rubric("REQUIREMENT:")
    what = f"the interface group ``{ctx.item['name']}``"
    try:
        parent = ctx.item.parent("interface-ingroup")
        parent_kind = get_kind(parent)
        ctx.content.wrap(f"The {ctx.mapper.get_link(parent)} "
                         f"{parent_kind} shall contain {what}.")
    except IndexError:
        ctx.content.wrap(f"There shall be {what}.")
    _add_text(ctx, "brief", "BRIEF DESCRIPTION")
    _add_text(ctx, "description", "DESCRIPTION")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_header_file(ctx: _Context) -> None:
    ctx.content.add_rubric("REQUIREMENT:")
    path = ctx.item["path"]
    container = ctx.item.parent('interface-placement')
    container_kind = get_kind(container)
    ctx.content.wrap(f"The {ctx.mapper.get_link(container)} {container_kind} "
                     f"shall provide the header file ``<{path}>``.")
    _add_text(ctx, "brief", "BRIEF DESCRIPTION")
    _add_code_block(ctx.content, f"#include <{path}>")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_forward_declaration(ctx: _Context) -> None:
    ctx.content.add_rubric("REQUIREMENT:")
    container = ctx.item.parent('interface-placement')
    container_kind = get_kind(container)
    ctx.content.wrap(
        f"The {ctx.mapper.get_link(container)} {container_kind} shall "
        "provide a forward declaration of "
        f"{ctx.mapper.get_link(ctx.item.parent('interface-target'))}.")
    _add_code_block(ctx.content, f"{forward_declaration(ctx.item)};")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_object(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_text(ctx, "brief", "BRIEF DESCRIPTION")
    definition = _SPACE.sub(
        " ", _item_definition(ctx.item, ctx.code_mapper, ctx.spec.enabled_set))
    _add_code_block(ctx.content, f"extern {definition};")
    _add_text(ctx, "description", "DESCRIPTION")
    _add_text(ctx, "notes", "NOTES")


def _document_typedef(ctx: _Context) -> None:
    _document_unspecified(ctx)
    _add_text(ctx, "brief", "BRIEF DESCRIPTION")
    definition = _SPACE.sub(
        " ", _item_definition(ctx.item, ctx.code_mapper, ctx.spec.enabled_set))
    _add_code_block(ctx.content, f"typedef {definition};")
    _add_text(ctx, "description", "DESCRIPTION")
    _add_text(ctx, "notes", "NOTES")
    _add_default_links(ctx)
    _add_validations(ctx)


def _document_validation_by_analysis(ctx: _Context) -> None:
    _add_text(ctx, "text", "ANALYSIS")
    _add_validated_items(ctx)


def _document_validation_by_inspection(ctx: _Context) -> None:
    _add_text(ctx, "text", "INSPECTION")
    _add_validated_items(ctx)


def _document_validation_by_review_of_design(ctx: _Context) -> None:
    _add_text(ctx, "text", "REVIEW OF DESIGN")
    _add_validated_items(ctx)


_ITEM_DOCUMENTER = {
    "constraint": _document_constraint,
    "glossary/group": _document_requirement,
    "glossary/term": _document_requirement,
    "interface/appl-config-group": _document_acfg_group,
    "interface/appl-config-option/feature": _document_acfg_option,
    "interface/appl-config-option/feature-enable": _document_acfg_option,
    "interface/appl-config-option/initializer": _document_acfg_option,
    "interface/appl-config-option/integer": _document_acfg_option,
    "interface/define": _document_define,
    "interface/domain": _document_domain,
    "interface/enum": _document_enumeration,
    "interface/enumerator": _document_enumerator,
    "interface/forward-declaration": _document_forward_declaration,
    "interface/function": _document_directive,
    "interface/group": _document_group,
    "interface/header-file": _document_header_file,
    "interface/macro": _document_directive,
    "interface/register-block": _document_register_block,
    "interface/struct": _document_compound,
    "interface/typedef": _document_typedef,
    "interface/union": _document_compound,
    "interface/unspecified-define": _document_unspecified_define,
    "interface/unspecified-enum": _document_unspecified_type,
    "interface/unspecified-enumerator": _document_unspecified_enumerator,
    "interface/unspecified-function": _document_unspecified_function,
    "interface/unspecified-group": _document_unspecified_group,
    "interface/unspecified-header-file": _document_header_file,
    "interface/unspecified-macro": _document_unspecified_function,
    "interface/unspecified-object": _document_unspecified_object,
    "interface/unspecified-struct": _document_unspecified_type,
    "interface/unspecified-typedef": _document_unspecified_typedef,
    "interface/unspecified-union": _document_unspecified_type,
    "interface/variable": _document_object,
    "memory-benchmark": _document_requirement,
    "requirement/functional/action": _document_action_requirement,
    "requirement/functional/capability": _document_requirement,
    "requirement/functional/interface-define-not-defined":
    _document_requirement,
    "requirement/functional/function": _document_requirement,
    "requirement/non-functional/design": _document_requirement,
    "requirement/non-functional/design-group": _document_requirement,
    "requirement/non-functional/design-target": _document_requirement,
    "requirement/non-functional/interface": _document_requirement,
    "requirement/non-functional/interface-requirement": _document_requirement,
    "requirement/non-functional/performance": _document_requirement,
    "requirement/non-functional/performance-runtime": _document_perf_runtime,
    "requirement/non-functional/performance-runtime-environment":
    _document_perf_runtime_env,
    "requirement/non-functional/quality": _document_requirement,
    "runtime-measurement-test": _document_requirement,
    "test-case": _document_requirement,
    "test-suite": _document_requirement,
    "validation/by-analysis": _document_validation_by_analysis,
    "validation/by-inspection": _document_validation_by_inspection,
    "validation/by-review-of-design": _document_validation_by_review_of_design,
}


def _status(item: Item) -> str:
    if not item.view["pre-qualified"]:
        return "not pre-qualified"
    if item.view["validated"]:
        return "validated"
    return "**not validated**"


def _validation_role(item: Item, role: str) -> str:
    if role.startswith("validation"):
        return role
    return f"{_status(item)} {role}"


def _validation_status(item: Item, role: str) -> str:
    if role.startswith("validation"):
        return role
    return f"{_status(item)} {role}"


def _is_include_enabled(enabled_set: EnabledSet, link: Link) -> bool:
    return is_enabled(enabled_set, link["enabled-by"])


def _add_validations(ctx: _Context) -> None:
    status = _status(ctx.item)
    kind = get_kind(ctx.item)
    validations = ctx.item.view["validation-dependencies"]
    if len(validations) == 0:
        ctx.content.add_rubric("VALIDATION:")
        ctx.content.add(f"This {kind} is {status}.")
    elif len(validations) == 1:
        ctx.content.add_rubric("VALIDATION:")
        validation = validations[0]
        item_2 = ctx.item.cache[validation[0]]
        link = ctx.mapper.get_link(ctx.item.cache[validation[0]],
                                   document_key="test-plan")
        if ctx.item == item_2:
            ctx.content.add(f"This {kind} is validated by a {validation[1]} "
                            f"specified by {link}.")
        else:
            ctx.content.add(
                f"This {status} {kind} is validated by the "
                f"{_validation_status(item_2, validation[1])} {link}.")
    else:
        ctx.content.add_rubric("VALIDATIONS:")
        items: list[str] = []
        for validation in validations:
            item_2 = ctx.item.cache[validation[0]]
            link = ctx.mapper.get_link(item_2, document_key="test-plan")
            items.append(
                f"{link} ({_validation_status(item_2, validation[1])})")
        ctx.content.add_list(
            items, f"The validation of this {status} {kind} "
            "depends on the following items:")


_VALIDATION_APPROACH = """The specification is a tree of specification items.
The root of the specification tree is ${/req/root:/spec}.  For each requirement
and interface a validation status can be determined.  An interface is *not
pre-qualified* if and only if at least one of the following conditions is met:

* *N1*: It has the ${/acfg/constraint/option-not-pre-qualified:/spec} usage
  constraint.

* *N2*: It has the ${/constraint/constant-not-pre-qualified:/spec} usage
  constraint.

* *N3*: It has the ${/constraint/directive-not-pre-qualified:/spec} usage
  constraint.

* *N4*: It is an interface container and all interfaces placed into this
  container are *not pre-qualified*.

An item is *validated* if and only if at least one of the following conditions
is met:

* *V1*: It has at least one not *not pre-qualified* refinement and all its
  refinements are *validated* or *not pre-qualified*.

* *V2*: It is a validation by test and at least one test result is available
  for this test and there are no unexpected test failures in the test results
  for this test.

* *V3*: It is a validation by analysis, inspection, or review of design.

* *V4*: It is a glossary term and its a member of a glossary group.

* *V5*: It is a constraint and it is a refinement of
  ${/req/usage-constraints:/spec}.

* *V6*: It is a design target and at least one test result is available for
  this target and there are no unexpected test failures in the test results for
  this target.

An item which is neither *validated* nor *not pre-qualified* is *not
validated*.  To check that all items are validated it is sufficient to check
the status of the root item: SPEC_ROOT.

Because of condition *V1* it is important to also consider the *not
pre-qualified* items in the validation procedure.  Some of the not
pre-qualified interfaces are fully specified with documentation entries.  They
are used to generate the ${/glossary/api:/term} header files and documentation.
In the generated API header files, there is a mix of pre-qualified and not
pre-qualified interfaces.  The not pre-qualified interfaces have no functional
specification and their implementation is removed from the pre-qualified
libraries.  For example, the use of a not pre-qualified function, would lead to
unresolved symbols at application link time.  In the user documentation, they
are marked as not pre-qualified through the corresponding usage constraint.

There are the following roles of refinement and validation items:

* *refinement*: The item is a requirement which refines a more general
  requirement.

* *group member*: The item is a member of an interface group.  Interfaces are
  organized in interface groups.  The interface groups define the software
  architecture and detailed design components.

* *interface placement*: The item is placed into an interface container.  For
  example, header files are interface containers.

* *interface function*: The item is a functional requirement which defines a
  function of an interface.

* *interface enumerator*: The item is an enumerator of an enumeration
  interface.

* *function implementation*: The item is the specification of a function used
  to implement interface functions.

* *validation by test*: The item is a test case which validates a requirement.
  The test case status is listed along the role in parenthesis where ``P``
  indicates a passed test case, ``F`` indicates an unexpectedly failed test
  case, and ``X`` indicates an expectedly failed test case.  For each test
  result, a status is listed.

* *validation by analysis*, *validation by inspection*, *validation by review
  of design*: The item is an analysis, inspection, or review of design which
  validates a requirement.

The table below lists for each requirement and interface related to the current
document, the validation status (*validated*, *not pre-qualified*, or *not
validated*), the associated refinement or validation items, and the role of the
refinement or validation item.  The table is a linearization of the
specification tree.  Parent items are on the left hand side.  Child items are
on the right hand side.  The table is ordered by the tree depth of parent items
starting with the root item.  Some items contain a functional or performance
specification along the associated validation test code.  These items show up
both as parent and child item in the same row where the child item has a
validation by test role.  The *not pre-qualified* interfaces (with the
exception of interface groups and containers) have no associated refinements or
validations, so the corresponding table entries are N/A."""


class SpecDocumentBuilder(DocumentBuilder):
    """ Builds documents presenting specification items. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        spec = self.input("spec")
        assert isinstance(spec, RTEMSItemCache)
        self.spec = spec
        try:
            spec_compare_registry = self.input("spec-compare-registry")
            assert isinstance(spec_compare_registry, CompareSpecsRegistry)
        except KeyError:
            spec_compare_registry = None
        self.spec_compare_registry = spec_compare_registry
        my_type = self.item.type
        self.mapper.add_get_value(f"{my_type}:/validation-verification",
                                  self._validation_verification)

    def get_items_of_document(self) -> list[Item]:
        """ Get the items of this document. """
        return self.spec.get_related_interfaces_and_requirements()

    def add_item_changes(self, content: SphinxContent, item: Item) -> None:
        """ Add the item changes to the content. """
        spec_compare_registry = self.spec_compare_registry
        if spec_compare_registry is not None:
            content.add_rubric("CHANGES:")
            spec_compare_registry.add_item_changes(content, item.uid)

    def add_item(self, content: SphinxContent, item: Item) -> None:
        """ Add the item documentation to the content. """
        content.register_license_and_copyrights_of_item(item)
        with self.mapper.scope(item):
            with content.directive("raw", "latex"):
                content.add("\\clearpage")
            with content.section(item.spec, label=spec_label(item)):
                _ITEM_DOCUMENTER[item.type](_Context(content, item,
                                                     self.mapper,
                                                     CodeMapper(item),
                                                     self.spec,
                                                     self.file_path))
                self.add_item_changes(content, item)

    def _add_validation_table(self, content: SphinxContent) -> None:
        """ Add the document item validation table to the content. """
        get_link = self.mapper.get_link
        item_cache = self.item.cache
        items = self.get_items_of_document()
        root_item = item_cache["/req/root"]
        if root_item in items:
            spec_root = ":ref:`spec:/req/root </req/root>`"
        else:
            spec_root = get_link(root_item)
        content.add(
            self.mapper.substitute(
                _VALIDATION_APPROACH.replace("SPEC_ROOT", spec_root)))
        rows: list[tuple[str | int,
                         ...]] = [("Interface / Requirement", "Status",
                                   "Refinement / Validation", "Role")]
        for item in sorted(items,
                           key=lambda x:
                           (len(x.view["order"]), x.view["order"])):
            req: str | int = f"_`{item.uid}`"
            status: str | int = _status(item)
            for validation in item.view["validation-dependencies"]:
                item_2 = item_cache[validation[0]]
                role = _validation_role(item_2, validation[1])
                if role == "validation by test":
                    role = f"{role} {get_test_result_status(item_2)}"
                if item != item_2 and item_2 in items:
                    item_2_ref = f"`{item_2.uid}`_"
                elif item_2.type == "memory-benchmark":
                    item_2_ref = (f":ref:`{item_2.uid} "
                                  f"<BenchmarkSpec{item_2.ident}>`")
                else:
                    item_2_ref = get_link(item_2,
                                          document_key="test-plan").replace(
                                              "spec:", "")
                rows.append((req, status, item_2_ref, role))
                req = COL_SPAN
                status = COL_SPAN
            if isinstance(req, str) and req:
                rows.append((req, status, "N/A",
                             item.view.get("validation-status", "N/A")))
        content.add_grid_table(rows, [32, 14, 32, 22], font_size=-4)

    def _validation_verification(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level)
            self._add_validation_table(content)
        return content.join()
